from __future__ import annotations

import collections.abc
import dataclasses
import datetime as dt
import json
import random
import typing
import warnings
from functools import cached_property
from typing import TYPE_CHECKING, Any, Callable, List, Literal, Mapping, Optional, Sequence, Tuple, TypeVar, Union
from urllib.parse import urlparse

import grpc
import grpc.experimental
from google.protobuf import empty_pb2, timestamp_pb2

from chalk import DataFrame, EnvironmentId, chalk_logger
from chalk._gen.chalk.auth.v1.agent_pb2 import CustomClaim
from chalk._gen.chalk.auth.v1.permissions_pb2 import Permission
from chalk._gen.chalk.common.v1 import online_query_pb2, upload_features_pb2
from chalk._gen.chalk.common.v1.online_query_pb2 import GenericSingleQuery, UploadFeaturesBulkRequest
from chalk._gen.chalk.engine.v1 import query_server_pb2
from chalk._gen.chalk.engine.v1.query_server_pb2_grpc import QueryServiceStub
from chalk._gen.chalk.graph.v1.graph_pb2 import Graph
from chalk._gen.chalk.protosql.v1.sql_service_pb2 import ExecuteSqlQueryRequest, PlanSqlQueryRequest
from chalk._gen.chalk.protosql.v1.sql_service_pb2_grpc import SqlServiceStub
from chalk._gen.chalk.server.v1.auth_pb2_grpc import AuthServiceStub
from chalk._gen.chalk.server.v1.deploy_pb2 import (
    CreateBranchFromSourceDeploymentRequest,
    CreateBranchFromSourceDeploymentResponse,
)
from chalk._gen.chalk.server.v1.deploy_pb2_grpc import DeployServiceStub
from chalk._gen.chalk.server.v1.graph_pb2 import (
    GetCodegenFeaturesFromGraphRequest,
    GetCodegenFeaturesFromGraphResponse,
    GetGraphRequest,
    GetGraphResponse,
    PythonVersion,
)
from chalk._gen.chalk.server.v1.graph_pb2_grpc import GraphServiceStub
from chalk._gen.chalk.server.v1.team_pb2 import (
    CreateServiceTokenRequest,
    CreateServiceTokenResponse,
    ListServiceTokensRequest,
    ListServiceTokensResponse,
)
from chalk._gen.chalk.server.v1.team_pb2_grpc import TeamServiceStub
from chalk.client import ChalkAuthException, FeatureReference
from chalk.client.client_impl import _validate_context_dict  # pyright: ignore[reportPrivateUsage]
from chalk.client.models import (
    BulkOnlineQueryResponse,
    BulkOnlineQueryResult,
    BulkUploadFeaturesResult,
    CreateBranchResponse,
    OnlineQuery,
    OnlineQueryResponse,
    UploadFeaturesResponse,
)
from chalk.client.serialization.protos import ChalkErrorConverter, OnlineQueryConverter, UploadFeaturesBulkConverter
from chalk.config.auth_config import TokenConfig, load_token
from chalk.features import live_updates
from chalk.features._encoding.inputs import GRPC_ENCODE_OPTIONS, InputEncodeOptions, recursive_encode_bulk_inputs
from chalk.features._encoding.json import FeatureEncodingOptions
from chalk.features._encoding.outputs import encode_outputs
from chalk.features.feature_set import is_feature_set_class
from chalk.features.tag import DeploymentId
from chalk.importer import CHALK_IMPORT_FLAG
from chalk.parsed._proto.utils import datetime_to_proto_timestamp, value_to_proto
from chalk.utils import df_utils
from chalk.utils.df_utils import record_batch_to_arrow_ipc
from chalk.utils.grpc import AuthenticatedChalkClientInterceptor, TokenRefresher, UnauthenticatedChalkClientInterceptor

if TYPE_CHECKING:
    from pyarrow import RecordBatch, Table

CHALK_GRPC_TRACE_ID_HEADER: str = "x-chalk-trace-id"


@dataclasses.dataclass
class ParsedUri:
    uri_without_scheme: str
    use_tls: bool


def get_trace_id_from_response(call: grpc.Call) -> Optional[str]:
    for k, v in call.trailing_metadata() or []:
        if k == CHALK_GRPC_TRACE_ID_HEADER:
            if isinstance(v, bytes):
                v = v.decode("utf-8")
            assert isinstance(v, str)  # for pyright
            return v
    return None


def _canonicalize_headers(
    headers: None | Sequence[tuple[str, str | bytes]] | Mapping[str, str | bytes],
) -> tuple[tuple[str, str | bytes], ...]:
    if headers is None:
        return ()
    # NOTE: metadata _keys_ must be lowercase
    if isinstance(headers, collections.abc.Mapping):
        return tuple((k.lower(), v) for (k, v) in headers.items())
    return tuple((k.lower(), v) for (k, v) in headers)


def get_features_feather_bytes(
    inputs: "Mapping[FeatureReference, Sequence[Any]] | DataFrame | Table | RecordBatch",
    options: InputEncodeOptions,
    compression: Literal["lz4", "zstd", "uncompressed"] = "lz4",
) -> bytes:
    import pyarrow as pa

    if isinstance(inputs, Mapping):
        inputs, _ = recursive_encode_bulk_inputs(inputs, options=options)

    if isinstance(inputs, DataFrame):
        inputs_table: pa.Table = inputs.to_pyarrow()
        input_batch = df_utils.pa_table_to_recordbatch(inputs_table)
    elif isinstance(inputs, pa.Table):
        input_batch = df_utils.pa_table_to_recordbatch(inputs)
    elif isinstance(inputs, pa.RecordBatch):
        input_batch = inputs
    else:
        encoded_inputs = {str(k): v for k, v in inputs.items()}
        input_batch = pa.RecordBatch.from_pydict(encoded_inputs)
    inputs_bytes = record_batch_to_arrow_ipc(input_batch, compression=compression)
    return inputs_bytes


def _parse_uri_for_engine(query_server_uri: str) -> ParsedUri:
    """
    If the scheme is provided, base TLS off of that (http = no tls, https = tls)
    If there is no scheme, default to TLS EXCEPT for localhost/private-vpc uris.
    """
    url_parsed = urlparse(query_server_uri)
    if url_parsed.scheme == "http":
        use_tls = False
    elif url_parsed.scheme == "https":
        use_tls = True
    elif url_parsed.scheme == "" and any(query_server_uri.startswith(pfx) for pfx in ["localhost", "127.0.0.1", "10."]):
        use_tls = False
    else:
        use_tls = True
    uri_without_scheme = query_server_uri.removeprefix(url_parsed.scheme + "://")
    return ParsedUri(uri_without_scheme=uri_without_scheme, use_tls=use_tls)


channel_options: List[tuple[str, str | int]] = [
    ("grpc.max_send_message_length", 1024 * 1024 * 100),  # 100MB
    ("grpc.max_receive_message_length", 1024 * 1024 * 100),  # 100MB
    # https://grpc.io/docs/guides/performance/#python
    (grpc.experimental.ChannelOptions.SingleThreadedUnaryStream, 1),
    (
        "grpc.service_config",
        json.dumps(
            {
                "methodConfig": [
                    {
                        "name": [{}],
                        "maxAttempts": 5,
                        "initialBackoff": "0.1s",
                        "maxBackoff": "1s",
                        "backoffMultiplier": 2,
                        "retryableStatusCodes": ["UNAVAILABLE"],
                    }
                ]
            }
        ),
    ),
]


T = TypeVar("T")
U = TypeVar("U")


class StubProvider:
    def close(self):
        if self._server_channel is not None:
            self._server_channel.close()
        if self._engine_channel is not None:
            self._engine_channel.close()

    @cached_property
    def deploy_stub(self):
        if self._server_channel is None:
            raise RuntimeError("Unable to connect to API server.")
        return DeployServiceStub(self._server_channel)

    @cached_property
    def graph_stub(self):
        if self._server_channel is None:
            raise RuntimeError("Unable to connect to API server.")
        return GraphServiceStub(self._server_channel)

    @cached_property
    def team_stub(self):
        if self._server_channel is None:
            raise RuntimeError("Unable to connect to API server.")
        return TeamServiceStub(self._server_channel)

    @cached_property
    def query_stub(self) -> QueryServiceStub:
        if self._engine_channel is None:
            raise ValueError(
                "The GRPC engine service is not available. If you would like to set up a GRPC service, please contact Chalk."
            )
        return QueryServiceStub(self._engine_channel)

    @cached_property
    def sql_stub(self) -> SqlServiceStub:
        if self._engine_channel is None:
            raise ValueError(
                "The GRPC engine service is not available. If you would like to set up a GRPC service, please contact Chalk."
            )
        return SqlServiceStub(self._engine_channel)

    def __init__(
        self,
        token_config: TokenConfig,
        query_server: str | None = None,
        deployment_tag: str | None = None,
        skip_api_server: bool = False,
        additional_headers: List[tuple[str, str]] | None = None,
    ):
        super().__init__()
        additional_headers_nonempty: List[tuple[str, str]] = [] if additional_headers is None else additional_headers
        token_refresher: TokenRefresher | None = None
        if skip_api_server:
            # Omits the auth handshake with the API server. Primarily for internal use/testing -- if used in production,
            # this client will simply fail to connect. If True then query_server must be provided & point to
            # `localhost/127.0.0.1`.
            if query_server is None:
                raise ValueError("If skipping API server auth, query_server URI must be provided.")
            elif not (query_server.startswith("localhost") or query_server.startswith("127.0.0.1")):
                warnings.warn(
                    "Skipping API server auth should only be enabled if query_server URI is localhost. It will fail to authenticate against a production engine."
                )
            self.environment_id = token_config.activeEnvironment
            if self.environment_id is None or self.environment_id == "":
                raise ValueError("No environment specified")
            self._server_channel: Optional[grpc.Channel] = None
        else:
            server_host: str = token_config.apiServer or "api.chalk.ai"
            for pfx in ("https://", "http://", "www."):
                server_host = server_host.removeprefix(pfx)

            _unauthenticated_server_channel: grpc.Channel = (
                grpc.insecure_channel(
                    target=server_host,
                    options=channel_options,
                )
                if server_host.startswith("localhost") or server_host.startswith("127.0.0.1")
                else grpc.secure_channel(
                    target=server_host,
                    credentials=grpc.ssl_channel_credentials(),
                    options=channel_options,
                )
            )

            self._auth_stub: AuthServiceStub = AuthServiceStub(
                grpc.intercept_channel(
                    _unauthenticated_server_channel,
                    UnauthenticatedChalkClientInterceptor(
                        server="go-api",
                        additional_headers=additional_headers_nonempty,
                    ),
                )
            )

            token_refresher = TokenRefresher(
                auth_stub=self._auth_stub,
                client_id=token_config.clientId,
                client_secret=token_config.clientSecret,
            )

            t = token_refresher.get_token()

            self.environment_id = token_config.activeEnvironment or t.primary_environment
            if not self.environment_id:
                raise ValueError("No environment specified")

            if self.environment_id not in t.environment_id_to_name:
                lower_env_id = self.environment_id.lower()
                valid = [eid for eid, ename in t.environment_id_to_name.items() if ename.lower() == lower_env_id]
                if len(valid) > 1:
                    raise ValueError(f"Multiple environments with name {self.environment_id}: {valid}")
                elif len(valid) == 0:
                    raise ValueError(f"No environment with name {self.environment_id}: {t.environment_id_to_name}")
                else:
                    self.environment_id = valid[0]

            self._server_channel: Optional[grpc.Channel] = grpc.intercept_channel(
                _unauthenticated_server_channel,
                AuthenticatedChalkClientInterceptor(
                    refresher=token_refresher,
                    server="go-api",
                    environment_id=self.environment_id,
                    additional_headers=additional_headers_nonempty,
                ),
            )

            query_server = query_server or t.grpc_engines.get(self.environment_id, None)
        engine_headers = additional_headers_nonempty + [("x-chalk-deployment-type", "engine-grpc")]
        if deployment_tag is not None:
            engine_headers += [("x-chalk-deployment-tag", deployment_tag)]
        interceptors: List[grpc.UnaryUnaryClientInterceptor] = [
            (
                AuthenticatedChalkClientInterceptor(
                    refresher=token_refresher,
                    environment_id=self.environment_id,
                    server="engine",
                    additional_headers=engine_headers,
                )
                if token_refresher is not None
                else UnauthenticatedChalkClientInterceptor(
                    server="engine",
                    additional_headers=engine_headers + [("x-chalk-env-id", self.environment_id)],
                )
            )
        ]

        self._engine_channel: Optional[grpc.Channel] = None
        if query_server is not None:
            parsed_uri = _parse_uri_for_engine(query_server_uri=query_server)
            self._engine_channel = grpc.intercept_channel(
                (
                    grpc.secure_channel(
                        target=parsed_uri.uri_without_scheme,
                        credentials=grpc.ssl_channel_credentials(),
                        options=channel_options,
                    )
                    if parsed_uri.use_tls
                    else grpc.insecure_channel(
                        target=parsed_uri.uri_without_scheme,
                        options=channel_options,
                    )
                ),
                *interceptors,
            )


class StubRefresher:
    def __init__(
        self,
        token_config: TokenConfig,
        query_server: str | None = None,
        deployment_tag: str | None = None,
        skip_api_server: bool = False,
        additional_headers: List[tuple[str, str]] | None = None,
    ):
        super().__init__()
        self._token_config = token_config
        self._query_server = query_server
        self._deployment_tag = deployment_tag
        self._skip_api_server = skip_api_server
        self._additional_headers = additional_headers
        self._stub = self._refresh_stub()

    def _refresh_stub(self) -> StubProvider:
        self._stub = StubProvider(
            token_config=self._token_config,
            query_server=self._query_server,
            deployment_tag=self._deployment_tag,
            skip_api_server=self._skip_api_server,
            additional_headers=self._additional_headers,
        )
        return self._stub

    def close(self):
        self._stub.close()

    def _retry_callable(self, fn: Callable[[T], U], get_service: Callable[[], T]) -> U:
        try:
            return fn(get_service())
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE and (
                "FD Shutdown" in (e.details() or "") or "GOAWAY" in (e.details() or "")
            ):
                chalk_logger.info("Detected FD shutdown; retrying connection: %s", e.details())
                old_stub = self._stub
                self._refresh_stub()
                old_stub.close()
                return fn(get_service())
            raise

    def call_deploy_stub(self, fn: Callable[[DeployServiceStub], T]) -> T:
        return self._retry_callable(fn, lambda: self._stub.deploy_stub)

    def call_graph_stub(self, fn: Callable[[GraphServiceStub], T]) -> T:
        return self._retry_callable(fn, lambda: self._stub.graph_stub)

    def call_team_stub(self, fn: Callable[[TeamServiceStub], T]) -> T:
        return self._retry_callable(fn, lambda: self._stub.team_stub)

    def call_query_stub(self, fn: Callable[[QueryServiceStub], T]) -> T:
        return self._retry_callable(fn, lambda: self._stub.query_stub)

    def call_sql_stub(self, fn: Callable[[SqlServiceStub], T]) -> T:
        return self._retry_callable(fn, lambda: self._stub.sql_stub)

    @property
    def environment_id(self) -> str | None:
        return self._stub.environment_id


class ChalkGRPCClient:
    """The `ChalkGRPCClient` is the primary Python interface for interacting with Chalk gRPC servers.

    You can use it to run online and offline queries targeted at the gRPC server, get data about the graph of
    features and resolvers deployed, and more."""

    def __init__(
        self,
        environment: EnvironmentId | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        api_server: str | None = None,
        deployment_tag: str | None = None,
        additional_headers: List[tuple[str, str]] | None = None,
        query_server: str | None = None,
        input_compression: typing.Literal["lz4", "zstd", "uncompressed"] = "lz4",
        **kwargs: Any,
    ):
        """Create a `ChalkGRPCClient` with the given credentials.

        Parameters
        ----------
        environment
            ID of the Chalk environment to connect to. If omitted, this is pulled from the current Chalk project config.
        client_id
            Client ID used to authenticate. If omitted, this is pulled from the current Chalk project config.
        client_secret
            Client secret used to authenticate. If omitted, this is pulled from the current Chalk project config.
        api_server
            URI of the Chalk API server used for authentication/metadata.
        additional_headers
            Additional client metadata to send with GRPC requests.
        query_server
            Hardcoded URI for Chalk query server, if available.
        deployment_tag
            Tag of the deployment to query. If omitted, the active deployment is used.
        """
        super().__init__()
        self._input_compression: typing.Literal["lz4", "zstd", "uncompressed"] = input_compression
        environment_id = kwargs.get("environment_id", None)
        if environment is not None and environment_id is not None:
            raise ValueError("Both environment and environment_id specified; only pass environment.")

        if environment_id is not None:
            environment = EnvironmentId(environment_id)

        # deprecating this.
        del environment_id

        if CHALK_IMPORT_FLAG.get() is True:
            raise RuntimeError(
                "Attempting to instantiate a Chalk client while importing source modules is forbidden. "
                + "Please exclude this file from import using your `.chalkignore` file "
                + "(see https://docs.chalk.ai/cli/apply), or wrap this query in a function that is not called upon import."
            )
        token_config = load_token(
            client_id=client_id,
            client_secret=client_secret,
            active_environment=environment,
            api_server=api_server,
            skip_cache=False,
        )
        if token_config is None:
            raise ChalkAuthException()

        self._stub_refresher = StubRefresher(
            token_config=token_config,
            query_server=query_server,
            deployment_tag=deployment_tag,
            additional_headers=additional_headers,
            skip_api_server=kwargs.get("_skip_api_server", False),
        )

    _INPUT_ENCODE_OPTIONS = GRPC_ENCODE_OPTIONS

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        self._stub_refresher.close()

    def ping_engine(self, num: Optional[int] = None) -> int:
        """
        Ping the engine to check if it is alive.

        Parameters
        ----------
        num
            A random number to send to the engine. If not provided, a random number is generated.
            This number will be returned as the response.

        Returns
        -------
        int
            The number sent to the engine.

        Examples
        --------
        >>> from chalk.client.client_grpc import ChalkGRPCClient
        >>> client = ChalkGRPCClient()
        >>> client.ping_engine(3)
        3
        """
        return self._stub_refresher.call_query_stub(
            lambda x: x.Ping(query_server_pb2.PingRequest(num=num if num is not None else random.randint(0, 999)))
        ).num

    def online_query(
        self,
        input: Union[Mapping[FeatureReference, Any], Any],
        output: Sequence[FeatureReference] = (),
        now: Optional[dt.datetime] = None,
        staleness: Optional[Mapping[FeatureReference, str]] = None,
        tags: List[str] | None = None,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        include_meta: bool = False,
        meta: Optional[Mapping[str, str]] = None,
        explain: bool = False,
        store_plan_stages: bool = False,
        value_metrics_tag_by_features: Optional[Sequence[FeatureReference]] = None,
        encoding_options: Optional[FeatureEncodingOptions] = None,
        required_resolver_tags: Optional[List[str]] = None,
        planner_options: Optional[Mapping[str, Any]] = None,
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str] | Sequence[tuple[str, str | bytes]] | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
    ) -> OnlineQueryResponse:
        """Compute features values using online resolvers.

        See https://docs.chalk.ai/docs/query-basics for more information.

        Parameters
        ----------
        input
            The features for which there are known values, mapped to those values.
            For example, `{User.id: 1234}`. Features can also be expressed as snakecased strings,
            e.g. `{"user.id": 1234}`
        output
            Outputs are the features that you'd like to compute from the inputs.
            For example, `[User.age, User.name, User.email]`.

            If an empty sequence, the output will be set to all features on the namespace
            of the query. For example, if you pass as input `{"user.id": 1234}`, then the query
            is defined on the `User` namespace, and all features on the `User` namespace
            (excluding has-one and has-many relationships) will be used as outputs.
        staleness
            Maximum staleness overrides for any output features or intermediate features.
            See https://docs.chalk.ai/docs/query-caching for more information.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        required_resolver_tags
            If specified, *all* required_resolver_tags must be present on a resolver for it to be
            considered eligible to execute.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        query_name
            The semantic name for the query you're making, for example, `"loan_application_model"`.
            Typically, each query that you make from your application should have a name.
            Chalk will present metrics and dashboard functionality grouped by 'query_name'.
        include_meta
            Returns metadata about the query execution under `OnlineQueryResult.meta`.
            This could make the query slightly slower.
            For more information, see https://docs.chalk.ai/docs/query-basics.
        explain
            Log the query execution plan. Requests using `explain=True` will be slower
            than requests using `explain=False`.

            If `True`, 'include_meta' will be set to `True` as well.
        store_plan_stages
            If `True`, the output of each of the query plan stages will be stored.
            This option dramatically impacts the performance of the query,
            so it should only be used for debugging.
        value_metrics_tag_by_features
            If your environment has feature value metrics enabled, this parameter specifies a list of features by which
            to tag these metrics. For example, if `value_metrics_tag_by_features=["user.category_id"]`, then the feature
             value metrics stored for this query will be tagged with the corresponding user's category_id.
        correlation_id
            You can specify a correlation ID to be used in logs and web interfaces.
            This should be globally unique, i.e. a `uuid` or similar. Logs generated
            during the execution of your query will be tagged with this correlation id.
        now
            The time at which to evaluate the query. If not specified, the current time will be used.
            This parameter is complex in the context of online_query since the online store
            only stores the most recent value of an entity's features. If `now` is in the past,
            it is extremely likely that `None` will be returned for cache-only features.

            This parameter is primarily provided to support:
                - controlling the time window for aggregations over cached has-many relationships
                - controlling the time wnidow for aggregations over has-many relationships loaded from an
                  external database

            If you are trying to perform an exploratory analysis of past feature values, prefer `offline_query`.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.

        Other Parameters
        ----------------
        meta
            Arbitrary `key:value` pairs to associate with a query.
        headers
            Additional headers to send with the request.
        planner_options
            Dictionary of additional options to pass to the Chalk query engine.
            Values may be provided as part of conversations with Chalk support
            to enable or disable specific functionality.
        request_timeout
            Float value indicating number of seconds that the request should wait before timing out
            at the network level. May not cancel resources on the server processing the query.

        Returns
        -------
        OnlineQueryResponse
            Wrapper around the output features and any query metadata
            and errors encountered while running the resolvers.

        Examples
        --------
        >>> from chalk.client.client_grpc import ChalkGRPCClient
        >>> result = ChalkGRPCClient().online_query(
        ...     input={
        ...         User.name: "Katherine Johnson"
        ...     },
        ...     output=[User.fico_score],
        ...     staleness={User.fico_score: "10m"},
        ... )
        """
        bulk_response = self._online_query_grpc_request(
            input=input,
            output=output,
            now=now,
            staleness=staleness,
            tags=tags,
            correlation_id=correlation_id,
            query_name=query_name,
            query_name_version=query_name_version,
            include_meta=include_meta,
            meta=meta,
            explain=explain,
            store_plan_stages=store_plan_stages,
            value_metrics_tag_by_features=value_metrics_tag_by_features,
            encoding_options=encoding_options,
            required_resolver_tags=required_resolver_tags,
            planner_options=planner_options,
            request_timeout=request_timeout,
            headers=headers,
            query_context=_validate_context_dict(query_context),
        )
        return OnlineQueryConverter.online_query_bulk_response_decode_to_single(bulk_response)

    def _online_query_grpc_request(
        self,
        *,
        input: Union[Mapping[FeatureReference, Any], Any],
        output: Sequence[FeatureReference] = (),
        now: Optional[dt.datetime] = None,
        staleness: Optional[Mapping[FeatureReference, str]] = None,
        tags: List[str] | None = None,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        include_meta: bool = False,
        meta: Optional[Mapping[str, str]] = None,
        explain: bool = False,
        store_plan_stages: bool = False,
        value_metrics_tag_by_features: Optional[Sequence[FeatureReference]] = None,
        encoding_options: Optional[FeatureEncodingOptions] = None,
        required_resolver_tags: Optional[List[str]] = None,
        planner_options: Optional[Mapping[str, Any]] = None,
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str] | Sequence[tuple[str, str | bytes]] | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | None = None,
    ) -> online_query_pb2.OnlineQueryBulkResponse:
        request = self._make_query_bulk_request(
            input={k: [v] for k, v in input.items()},
            output=output,
            now=[now] if now is not None else [],
            staleness=staleness or {},
            tags=tags or (),
            correlation_id=correlation_id,
            query_name=query_name,
            query_name_version=query_name_version,
            include_meta=include_meta,
            meta=meta or {},
            explain=explain,
            store_plan_stages=store_plan_stages,
            value_metrics_tag_by_features=value_metrics_tag_by_features,
            encoding_options=encoding_options,
            required_resolver_tags=required_resolver_tags or (),
            planner_options=planner_options or {},
            query_context=query_context,
        )
        return self._stub_refresher.call_query_stub(
            lambda x: x.OnlineQueryBulk(
                request,
                timeout=request_timeout,
                metadata=_canonicalize_headers(headers),
            )
        )

    def online_query_bulk(
        self,
        input: Union[Mapping[FeatureReference, Sequence[Any]], DataFrame],
        output: Sequence[FeatureReference] = (),
        now: Optional[Sequence[dt.datetime]] = None,
        staleness: Optional[Mapping[FeatureReference, str]] = None,
        tags: Optional[List[str]] = None,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        include_meta: bool = False,
        meta: Optional[Mapping[str, str]] = None,
        explain: bool = False,
        store_plan_stages: bool = False,
        encoding_options: Optional[FeatureEncodingOptions] = None,
        required_resolver_tags: Optional[List[str]] = None,
        value_metrics_tag_by_features: Optional[Sequence[FeatureReference]] = None,
        planner_options: Optional[Mapping[str, Union[str, int, bool]]] = None,
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str | bytes] | Sequence[tuple[str, str | bytes]] | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
    ) -> BulkOnlineQueryResult:
        response, call = self._online_query_bulk_grpc_request(
            input=input,
            output=output,
            now=now,
            staleness=staleness,
            tags=tags,
            correlation_id=correlation_id,
            query_name=query_name,
            query_name_version=query_name_version,
            include_meta=include_meta,
            meta=meta,
            explain=explain,
            store_plan_stages=store_plan_stages,
            value_metrics_tag_by_features=value_metrics_tag_by_features,
            encoding_options=encoding_options,
            required_resolver_tags=required_resolver_tags,
            planner_options=planner_options,
            request_timeout=request_timeout,
            headers=headers,
            query_context=_validate_context_dict(query_context),
        )
        return OnlineQueryConverter.online_query_bulk_response_decode(
            response, trace_id=get_trace_id_from_response(call)
        )

    def _online_query_bulk_grpc_request(
        self,
        *,
        input: Union[Mapping[FeatureReference, Sequence[Any]], DataFrame],
        output: Sequence[FeatureReference] = (),
        now: Optional[Sequence[dt.datetime]] = None,
        staleness: Optional[Mapping[FeatureReference, str]] = None,
        tags: Optional[List[str]] = None,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        include_meta: bool = False,
        meta: Optional[Mapping[str, str]] = None,
        explain: bool = False,
        store_plan_stages: bool = False,
        value_metrics_tag_by_features: Optional[Sequence[FeatureReference]] = None,
        encoding_options: Optional[FeatureEncodingOptions] = None,
        required_resolver_tags: Optional[List[str]] = None,
        planner_options: Optional[Mapping[str, Union[str, int, bool]]] = None,
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str | bytes] | Sequence[tuple[str, str | bytes]] | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | None = None,
    ) -> Tuple[online_query_pb2.OnlineQueryBulkResponse, grpc.Call]:
        """Returns the raw GRPC response and metadata"""
        request = self._make_query_bulk_request(
            input=input,
            output=output,
            now=now or (),
            staleness=staleness or {},
            tags=tags or (),
            correlation_id=correlation_id,
            query_name=query_name,
            query_name_version=query_name_version,
            include_meta=include_meta,
            meta=meta or {},
            explain=explain,
            store_plan_stages=store_plan_stages,
            value_metrics_tag_by_features=value_metrics_tag_by_features,
            encoding_options=encoding_options,
            required_resolver_tags=required_resolver_tags or (),
            planner_options=planner_options or {},
            query_context=query_context,
        )
        return self._stub_refresher.call_query_stub(
            lambda x: x.OnlineQueryBulk.with_call(
                request,
                timeout=request_timeout,
                metadata=_canonicalize_headers(headers),
            )
        )

    def upload_features_bulk(
        self,
        inputs: "Union[Mapping[FeatureReference, Sequence[Any]], DataFrame, Table, RecordBatch]",
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str] | Sequence[tuple[str, str | bytes]] | None = None,
    ) -> BulkUploadFeaturesResult:
        request = UploadFeaturesBulkRequest(
            inputs_feather=get_features_feather_bytes(inputs, self._INPUT_ENCODE_OPTIONS),
        )
        response, call = self._stub_refresher.call_query_stub(
            lambda x: x.UploadFeaturesBulk.with_call(
                request,
                timeout=request_timeout,
                metadata=_canonicalize_headers(headers),
            )
        )
        return UploadFeaturesBulkConverter.upload_features_bulk_response_decode(
            response,
            trace_id=get_trace_id_from_response(call),
        )

    def upload_features(
        self,
        inputs: "Union[Mapping[FeatureReference, Sequence[Any]], DataFrame, Table, RecordBatch]",
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str] | Sequence[tuple[str, str | bytes]] | None = None,
    ) -> UploadFeaturesResponse:
        """Upload data to Chalk to be inserted into the online & offline stores.

        Parameters
        ----------
        inputs
            Input data can be in one of two formats:
            1. A mapping from a feature or feature name to a list of values:
               `{Transaction.id: ["a", "b", "c"], Transaction.amount: [100.0,200.0,300.0]}`
            2. A tabular format such as arrow Table/RecordBatch, polars Dataframe, chalk.DataFrame
               where each column corresponds to a feature.
        headers
            Additional headers to send with the request.
        request_timeout
            Float value indicating number of seconds that the request should wait before timing out
            at the network level. May not cancel resources on the server processing the query.

        Returns
        -------
        UploadFeaturesResponse
            which contains a list of errors if any occurred.
        """
        request = upload_features_pb2.UploadFeaturesRequest(
            inputs_table=get_features_feather_bytes(inputs, self._INPUT_ENCODE_OPTIONS)
        )
        response, call = self._stub_refresher.call_query_stub(
            lambda x: x.UploadFeatures.with_call(
                request, timeout=request_timeout, metadata=_canonicalize_headers(headers)
            )
        )
        trace_id = get_trace_id_from_response(call)
        py_errors = [ChalkErrorConverter.chalk_error_decode(err) for err in response.errors]
        return UploadFeaturesResponse(errors=py_errors, trace_id=trace_id)

    def multi_query(
        self,
        queries: List[OnlineQuery],
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        include_meta: bool = False,
        meta: Optional[Mapping[str, str]] = None,
        explain: bool = False,
        store_plan_stages: bool = False,
        value_metrics_tag_by_features: Optional[Sequence[FeatureReference]] = None,
        encoding_options: Optional[FeatureEncodingOptions] = None,
        required_resolver_tags: Optional[Sequence[str]] = None,
        planner_options: Optional[Mapping[str, Any]] = None,
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str] | Sequence[tuple[str, str | bytes]] | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
    ) -> BulkOnlineQueryResponse:
        """Execute a series of independent requests in parallel."""
        requests: List[GenericSingleQuery] = []
        for query in queries:
            # NOTE: This assumed every request is a 'bulk' request.
            if value_metrics_tag_by_features is not None:
                query_vmtbf = value_metrics_tag_by_features
            else:
                query_vmtbf = query.value_metrics_tag_by_features
            request = self._make_query_bulk_request(
                input=query.input,
                output=query.output,
                now=(),
                staleness=query.staleness or {},
                tags=query.tags or (),
                correlation_id=correlation_id,
                query_name=query_name,
                query_name_version=query_name_version,
                include_meta=include_meta,
                meta=meta or {},
                explain=explain,
                store_plan_stages=store_plan_stages,
                value_metrics_tag_by_features=query_vmtbf,
                encoding_options=encoding_options,
                required_resolver_tags=required_resolver_tags or (),
                planner_options=query.planner_options or planner_options or {},
                query_context=query_context,
            )
            requests.append(GenericSingleQuery(bulk_request=request))
        response, call = self._stub_refresher.call_query_stub(
            lambda x: x.OnlineQueryMulti.with_call(
                online_query_pb2.OnlineQueryMultiRequest(
                    queries=requests,
                ),
                timeout=request_timeout,
                metadata=_canonicalize_headers(headers),
            )
        )
        return OnlineQueryConverter.online_query_multi_response_decode(
            response, trace_id=get_trace_id_from_response(call)
        )

    def _get_python_codegen(
        self, branch: str | None = None, deployment_id: str | None = None
    ) -> GetCodegenFeaturesFromGraphResponse:
        """Execute a series of independent requests in parallel."""
        import sys

        python_verison = sys.version_info

        resp: GetCodegenFeaturesFromGraphResponse = self._stub_refresher.call_graph_stub(
            lambda x: x.GetCodegenFeaturesFromGraph(
                GetCodegenFeaturesFromGraphRequest(
                    branch=branch,
                    deployment_id=deployment_id,
                    python_version=PythonVersion(
                        major=python_verison.major, minor=python_verison.minor, patch=python_verison.micro
                    ),
                )
            )
        )
        return resp

    def _create_branch(
        self, branch_name: str, source_branch_name: str | None = None, source_deployment_id: str | None = None
    ) -> CreateBranchResponse:
        """Create a branch from either a source deployment id, a source branch name, or the mainline deployment."""
        resp: CreateBranchFromSourceDeploymentResponse = self._stub_refresher.call_deploy_stub(
            lambda x: x.CreateBranchFromSourceDeployment(
                CreateBranchFromSourceDeploymentRequest(
                    branch_name=branch_name,
                    source_branch_name=source_branch_name,
                    source_deployment_id=source_deployment_id,
                    current_mainline_deployment=empty_pb2.Empty()
                    if source_branch_name is None and source_deployment_id is None
                    else None,
                )
            )
        )
        return CreateBranchResponse(
            branch_already_exists=resp.branch_already_exists,
            errors=[ChalkErrorConverter.chalk_error_decode(err) for err in resp.deployment_errors],
        )

    def _make_query_bulk_request(
        self,
        input: Mapping[FeatureReference, Sequence[Any]] | DataFrame,
        output: Sequence[FeatureReference],
        now: Sequence[dt.datetime],
        staleness: Mapping[FeatureReference, str],
        tags: Sequence[str],
        correlation_id: str | None,
        query_name: str | None,
        query_name_version: str | None,
        include_meta: bool,
        meta: Mapping[str, str],
        explain: bool,
        store_plan_stages: bool,
        value_metrics_tag_by_features: Optional[Sequence[FeatureReference]],
        encoding_options: FeatureEncodingOptions | None,
        required_resolver_tags: Sequence[str],
        planner_options: Mapping[str, str | int | bool],
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None,
    ) -> online_query_pb2.OnlineQueryBulkRequest:
        inputs_bytes = get_features_feather_bytes(
            input, self._INPUT_ENCODE_OPTIONS, compression=self._input_compression
        )
        encoded_outputs = encode_outputs(output)
        outputs = encoded_outputs.string_outputs
        # Currently assume every feature tag is just a fqn instead of a more complex expr.
        value_metrics_tags_encoded = encode_outputs(value_metrics_tag_by_features or []).string_outputs
        value_metrics_tags_proto = [online_query_pb2.OutputExpr(feature_fqn=o) for o in value_metrics_tags_encoded]

        now_proto: List[timestamp_pb2.Timestamp] = []
        for ts in now:
            if ts.tzinfo is None:
                ts = ts.astimezone(tz=dt.timezone.utc)
            now_proto.append(datetime_to_proto_timestamp(ts))

        staleness_encoded: dict[str, str] = {}
        for k, v in staleness.items():
            if is_feature_set_class(k):
                for f in k.features:
                    staleness_encoded[f.root_fqn] = v
            else:
                staleness_encoded[str(k)] = v

        context_options_dict: dict[str, Any] = {
            "store_plan_stages": store_plan_stages,
        }
        context_options_dict.update(**(planner_options or {}))
        context_options_proto = {k: value_to_proto(v) for k, v in context_options_dict.items()}
        query_context = _validate_context_dict(query_context)
        query_context_proto = {k: value_to_proto(v) for k, v in query_context.items()} if query_context else None
        return online_query_pb2.OnlineQueryBulkRequest(
            inputs_feather=inputs_bytes,
            outputs=[online_query_pb2.OutputExpr(feature_fqn=o) for o in outputs]
            + [online_query_pb2.OutputExpr(feature_expression=o) for o in encoded_outputs.feature_expressions_proto],
            now=now_proto,
            staleness=staleness_encoded,
            context=online_query_pb2.OnlineQueryContext(
                environment=self._stub_refresher.environment_id,
                tags=tags,
                required_resolver_tags=required_resolver_tags,
                correlation_id=correlation_id,
                query_name=query_name,
                query_name_version=query_name_version,
                options=context_options_proto,
                query_context=query_context_proto,
                value_metrics_tag_by_features=value_metrics_tags_proto,
                overlay_graph=live_updates.build_overlay_graph(),
            ),
            response_options=online_query_pb2.OnlineQueryResponseOptions(
                include_meta=include_meta,
                explain=online_query_pb2.ExplainOptions() if explain else None,
                encoding_options=online_query_pb2.FeatureEncodingOptions(
                    encode_structs_as_objects=encoding_options.encode_structs_as_objects if encoding_options else False
                ),
                metadata=meta,
            ),
            body_type=online_query_pb2.FEATHER_BODY_TYPE_RECORD_BATCHES,
        )

    def get_graph(self, deployment: DeploymentId | None = None) -> Graph:
        """Get the graph for a given deployment.

        Parameters
        ----------
        deployment
            The id of the Chalk deployment, or `None` to use the latest deployment.

        Returns
        -------
        Graph
            The graph for the given deployment.

        Examples
        --------
        >>> from chalk.client.client_grpc import ChalkGRPCClient
        >>> ChalkGRPCClient().get_graph()
        """
        resp: GetGraphResponse = self._stub_refresher.call_graph_stub(
            lambda x: x.GetGraph(GetGraphRequest(deployment_id=deployment))
        )
        return resp.graph

    def create_service_token(
        self,
        name: str,
        permissions: List[Permission],
        customer_claims: Mapping[str, List[str]] | None = None,
    ) -> CreateServiceTokenResponse:
        """Create a service token with a given set of permissions and claims.

        Parameters
        ----------
        name
            The name of your service token.
        permissions
            The permissions that you want your token to have.
        customer_claims
            The customer claims that you want your token to have.

        Returns
        -------
        CreateServiceTokenResponse
            A service token response, including a `client_id` and `client_secret` with
            the specified permissions and customer claims.

        Examples
        --------
        >>> from chalk.client import Permission
        >>> client = ChalkGRPCClient(client_id='test', client_secret='test_secret')
        >>> client.create_service_token(permissions=[Permission.PERMISSION_QUERY_ONLINE])
        """
        return self._stub_refresher.call_team_stub(
            lambda x: x.CreateServiceToken(
                CreateServiceTokenRequest(
                    name=name,
                    permissions=permissions,
                    customer_claims=(
                        None
                        if customer_claims is None
                        else [CustomClaim(key=key, values=values) for key, values in customer_claims.items()]
                    ),
                )
            )
        )

    def list_service_tokens(self) -> ListServiceTokensResponse:
        """Get all service tokens for the current environment.

        Returns
        -------
        ListServiceTokensResponse
            A list of service tokens for the current environment.

        Examples
        --------
        >>> from chalk.client import Permission
        >>> client = ChalkGRPCClient()
        >>> client.list_service_tokens()
        """
        return self._stub_refresher.call_team_stub(lambda x: x.ListServiceTokens(ListServiceTokensRequest()))

    def run_sql(self, sql: str):
        return self._stub_refresher.call_sql_stub(lambda x: x.ExecuteSqlQuery(ExecuteSqlQueryRequest(query=sql)))

    def explain_sql(self, sql: str):
        return self._stub_refresher.call_sql_stub(lambda x: x.PlanSqlQuery(PlanSqlQueryRequest(query=sql)))



from __future__ import annotations

import collections
import contextlib
import csv
import io
import logging
import os
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Deque, Dict, Iterable, List, Mapping, Optional, Sequence, Union, cast

import pyarrow as pa
import pyarrow.csv
from packaging.version import parse

from chalk.features import Feature
from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.query_execution_parameters import QueryExecutionParameters
from chalk.sql._internal.sql_source import (
    BaseSQLSource,
    SQLSourceKind,
    TableIngestMixIn,
    UnsupportedEfficientExecutionError,
    validate_dtypes_for_efficient_execution,
)
from chalk.sql.finalized_query import FinalizedChalkQuery, Finalizer
from chalk.sql.protocols import SQLSourceWithTableIngestProtocol
from chalk.utils.environment_parsing import env_var_bool
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.tracing import safe_add_metrics, safe_add_tags, safe_trace

if TYPE_CHECKING:
    import polars as pl
    from polars.type_aliases import PolarsTemporalType
    from sqlalchemy.engine import URL, Connection

try:
    import sqlalchemy as sa
except ImportError:
    sa = None

if sa is None:
    _supported_sqlalchemy_types_for_pa_csv_querying = ()
else:
    _supported_sqlalchemy_types_for_pa_csv_querying = (
        sa.BigInteger,
        sa.Boolean,
        sa.Float,
        sa.Integer,
        sa.String,
        sa.Text,
        sa.DateTime,
        sa.Date,
        sa.SmallInteger,
        sa.BIGINT,
        sa.BOOLEAN,
        sa.CHAR,
        sa.DATETIME,
        sa.FLOAT,
        sa.INTEGER,
        sa.SMALLINT,
        sa.TEXT,
        sa.TIMESTAMP,
        sa.VARCHAR,
    )

_logger = get_logger(__name__)


_PGHOST_NAME = "PGHOST"
_PGPORT_NAME = "PGPORT"
_PGDATABASE_NAME = "PGDATABASE"
_PGUSER_NAME = "PGUSER"
_PGPASSWORD_NAME = "PGPASSWORD"
_ENGINE_ARGUMENTS_NAME = "ENGINE_ARGUMENTS"


class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIngestProtocol):
    kind = SQLSourceKind.postgres

    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
        engine_args: Optional[Dict[str, Any]] = None,
        async_engine_args: Optional[Dict[str, Any]] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            import psycopg  # Used for the async driver
            import psycopg2  # Used for the sync driver
            from sqlalchemy.dialects import registry  # pyright: ignore
        except ImportError:
            raise missing_dependency_exception("chalkpy[postgresql]")
        del psycopg2  # unused
        del psycopg  # unused
        if "postgresql.psycopg" not in registry.impls:
            registry.register(
                "postgresql.psycopg", "chalk.sql._internal.integrations.psycopg3.psycopg_dialect", "dialect"
            )
        if "postgresql.psycopg_async" not in registry.impls:
            registry.register(
                "postgresql.psycopg_async", "chalk.sql._internal.integrations.psycopg3.psycopg_dialect", "dialect_async"
            )
        self.name = name
        self.host = host or load_integration_variable(
            integration_name=name, name=_PGHOST_NAME, override=integration_variable_override
        )
        self.port = (
            int(port)
            if port is not None
            else load_integration_variable(
                integration_name=name, name=_PGPORT_NAME, parser=int, override=integration_variable_override
            )
        )
        self.db = db or load_integration_variable(
            integration_name=name, name=_PGDATABASE_NAME, override=integration_variable_override
        )
        self.user = user or load_integration_variable(
            integration_name=name, name=_PGUSER_NAME, override=integration_variable_override
        )
        self.password = password or load_integration_variable(
            integration_name=name, name=_PGPASSWORD_NAME, override=integration_variable_override
        )
        self.ingested_tables: Dict[str, Any] = {}
        if engine_args is None:
            engine_args = {}
        if async_engine_args is None:
            async_engine_args = {}

        if name:
            engine_args_from_ui = self._load_env_engine_args(name, override=integration_variable_override)
            for k, v in engine_args_from_ui.items():
                engine_args.setdefault(k, v)
                async_engine_args.setdefault(k, v)

        chalk_default_engine_args = {
            "pool_size": 20,
            "max_overflow": 60,
            "pool_recycle": 90,
            "connect_args": {
                "keepalives": 1,
                "keepalives_idle": 30,
                "keepalives_interval": 10,
                "keepalives_count": 5,
            },
        }
        for k, v in chalk_default_engine_args.items():
            engine_args.setdefault(k, v)
            async_engine_args.setdefault(k, v)

        # We set the default isolation level to autocommit since the SQL sources are read-only, and thus
        # transactions are not needed
        # Setting the isolation level on the engine, instead of the connection, avoids
        # a DBAPI statement to reset the transactional level back to the default before returning the
        # connection to the pool
        engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
        async_engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
        BaseSQLSource.__init__(self, name=name, engine_args=engine_args, async_engine_args=async_engine_args)

    def get_sqlglot_dialect(self) -> str | None:
        return "postgres"

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="postgresql+psycopg2",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.db,
        )

    def async_local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="postgresql+psycopg_async",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.db,
        )

    def _get_parsed_table_via_polars(
        self,
        finalized_query: FinalizedChalkQuery,
        connection: Optional[Connection],
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        query_execution_parameters: QueryExecutionParameters,
    ) -> pyarrow.Table:
        with safe_trace("_get_parsed_table_via_polars"):
            import polars as pl

            buffer = self._get_result_csv_buffer(
                finalized_query,
                connection=connection,
                escape_char='"',
                query_execution_parameters=query_execution_parameters,
            )
            first_line = buffer.readline()
            column_names = list(
                csv.reader(
                    [first_line.decode("utf8")],
                    doublequote=True,
                    quoting=csv.QUOTE_MINIMAL,
                )
            )[0]
            buffer.seek(0)
            features = columns_to_features(column_names)

            parse_dtypes: Dict[str, pl.PolarsDataType] = {}
            boolean_columns: List[str] = []
            date_columns: List[str] = []
            datetime_col_to_dtype: Dict[str, pl.Datetime] = {}
            for col_name, feature in features.items():
                dtype = feature.converter.polars_dtype
                if dtype == pl.Boolean:  # pyright: ignore[reportUnnecessaryComparison]
                    # Parsing "t" and "f" as booleans fails.
                    # We will parse the column as a string, and then
                    # convert it to a boolean later.
                    parse_dtypes[col_name] = pl.Utf8
                    boolean_columns.append(col_name)
                elif dtype == pl.List:  # pyright: ignore[reportUnnecessaryComparison]
                    # We will parse the list as a string, and then
                    # convert it to a list later. Polars does not
                    # support parsing lists directly from CSV.
                    parse_dtypes[col_name] = pl.Utf8
                elif dtype == pl.Datetime:  # pyright: ignore[reportUnnecessaryComparison]
                    # Parse as str, convert to datetime later.
                    # It's important to parse datetimes as strings,
                    # because `polars.read_csv()` just turns
                    # unparsable datetimes into nulls, a huge no-no.
                    parse_dtypes[col_name] = pl.Utf8
                    # Make linter happy
                    dtype = cast(pl.Datetime, dtype)
                    datetime_col_to_dtype[col_name] = dtype
                elif dtype == pl.Date:  # pyright: ignore[reportUnnecessaryComparison]
                    # Parse as str, convert to date later.
                    # It's important to parse dates as strings,
                    # because `polars.read_csv()` just turns
                    # unparsable dates into nulls, a huge no-no.
                    parse_dtypes[col_name] = pl.Utf8
                    date_columns.append(col_name)
                else:
                    parse_dtypes[col_name] = dtype

            """
            CSV -> Polars -> Pyarrow table
            """
            # Previously we were using `pyarrow.csv.read_csv` but performance
            # degraded over time up till the next reboot.
            #
            # pl.read_csv(use_pyarrow=True) has the same performance degradation,
            # UNLESS a `dtypes` arg is provided.

            # 'dtypes' deprecated for 'schema_overrides' in polars 0.20+, but parameter renamed without breaking
            pl_table = pl.read_csv(buffer, dtypes=parse_dtypes)  # pyright: ignore[reportCallIssue]
            if boolean_columns:
                # DO NOT use map_dict. Causes a segfault when multiple uvicorn workers are handling
                # requests in parallel.
                boolean_when_mappings = [
                    pl.when(pl.col(bool_col).is_null())
                    .then(None)
                    .otherwise(pl.col(bool_col) == "t")
                    .cast(pl.Boolean)
                    .alias(bool_col)
                    for bool_col in boolean_columns
                ]
                pl_table = pl_table.with_columns(*boolean_when_mappings)
            if date_columns:
                for date_col in date_columns:
                    pl_table = _get_df_with_parsed_dt_column(
                        pl_table=pl_table,
                        col_name=date_col,
                        parser_formats=["%Y-%m-%d", "%D", "%m/%d/%Y"],
                        target_dtype=pl.Date,
                    )
            if datetime_col_to_dtype:
                from polars.type_aliases import TimeUnit

                for dt_col, dt_type in datetime_col_to_dtype.items():
                    unit_to_precision: Dict[TimeUnit, str] = {
                        "ms": "%.3f",
                        "us": "%.6f",
                        "ns": "%.9f",
                    }
                    precision = (
                        unit_to_precision[dt_type.time_unit]
                        if dt_type.time_unit and dt_type.time_unit in unit_to_precision
                        else "%.6f"
                    )
                    dt_format = f"%Y-%m-%d %H:%M:%S{precision}"
                    dt_format_with_tz = f"{dt_format}%#z"
                    datetime_parser_formats = [dt_format, dt_format_with_tz, "%Y-%m-%d"]
                    pl_table = _get_df_with_parsed_dt_column(
                        pl_table=pl_table,
                        col_name=dt_col,
                        parser_formats=datetime_parser_formats,
                        target_dtype=dt_type,
                    )

            parsed_table = pl_table.to_arrow()

            safe_add_metrics(
                {
                    "postgresl_polars_row_count": len(parsed_table),
                    "postgresl_polars_bytes": parsed_table.nbytes,
                }
            )
            return parsed_table

    def _get_result_csv_buffer(
        self,
        finalized_query: FinalizedChalkQuery,
        connection: Optional[Connection],
        escape_char: str,
        query_execution_parameters: QueryExecutionParameters,
    ) -> io.BytesIO:
        with safe_trace("_get_result_csv_buffer"):
            import psycopg2.sql

            stmt, positional_params, named_params = self.compile_query(finalized_query, paramstyle="named")
            assert len(positional_params) == 0, "should not have any positional params"

            # Convert the param style to python3-style {}, which is what psycopg2.sql.SQL.format expects
            # Forcing quote so the unquoted empty string will represent null values
            stmt = _reformat_compiled_query_string_for_csv_output(
                stmt,
                escape_char=escape_char,
                named_params=named_params,
            )
            if (
                env_var_bool("CHALK_SKIP_PG_DATETIME_ZONE_CAST")
                or query_execution_parameters.postgres.skip_datetime_timezone_cast
            ):
                with safe_trace("skip_datetime_timezone_cast"):
                    formatted_stmt = psycopg2.sql.SQL(stmt).format(
                        **{
                            k: psycopg2.sql.Literal(str(v)) if isinstance(v, datetime) else psycopg2.sql.Literal(v)
                            for (k, v) in named_params.items()
                        }
                    )
            else:
                formatted_stmt = psycopg2.sql.SQL(stmt).format(
                    **{k: psycopg2.sql.Literal(v) for (k, v) in named_params.items()}
                )
            with self.get_engine().connect() if connection is None else contextlib.nullcontext(connection) as cnx:
                dbapi = cnx.connection
                with dbapi.cursor() as cursor:
                    buffer = io.BytesIO()
                    with safe_trace("copy_expert"):
                        safe_add_tags({"query": formatted_stmt.as_string(cursor)[0:2000]})
                        cursor.copy_expert(formatted_stmt, buffer)
                        written_bytes = buffer.getbuffer().nbytes
                        safe_add_metrics({"postgresl_bytes": written_bytes})

            buffer.seek(0)

            return buffer

    def _execute_query_efficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[Connection],
        query_execution_parameters: QueryExecutionParameters,
    ) -> Iterable[pa.RecordBatch]:
        with safe_trace("postgres.execute_query_efficient"):
            from sqlalchemy.sql import Select

            if finalized_query.finalizer in (Finalizer.FIRST, Finalizer.ONE, Finalizer.ONE_OR_NONE):
                raise UnsupportedEfficientExecutionError(
                    (
                        f"Falling back to SQLAlchemy execution for finalizer '{finalized_query.finalizer.value}', "
                        "as it is faster for small results."
                    ),
                    log_level=logging.DEBUG,
                )

            if isinstance(finalized_query.query, Select):
                validate_dtypes_for_efficient_execution(
                    finalized_query.query, _supported_sqlalchemy_types_for_pa_csv_querying
                )

            assert len(finalized_query.temp_tables) == 0, "Should not create temp tables with postgres source"

            if query_execution_parameters.postgres.polars_read_csv:
                _logger.debug("postgres: doing force polars CSV reader")
                parsed_table = self._get_parsed_table_via_polars(
                    finalized_query, connection, columns_to_features, query_execution_parameters
                )
            else:
                buffer = self._get_result_csv_buffer(
                    finalized_query,
                    connection=connection,
                    escape_char="\\",
                    query_execution_parameters=query_execution_parameters,
                )
                buffer.seek(0)
                parsed_table = pyarrow.csv.read_csv(
                    buffer,
                    parse_options=pyarrow.csv.ParseOptions(
                        newlines_in_values=True,
                        escape_char="\\",
                        double_quote=False,
                    ),
                    convert_options=pyarrow.csv.ConvertOptions(
                        true_values=["t"],
                        false_values=["f"],
                        strings_can_be_null=True,
                        quoted_strings_can_be_null=False,
                    ),
                )

            features = columns_to_features(parsed_table.column_names)

            restricted_schema = pa.schema([pa.field(k, v.converter.pyarrow_dtype) for (k, v) in features.items()])
            parsed_table = parsed_table.select(list(features.keys()))
            parsed_table = parsed_table.cast(restricted_schema)
            parsed_table = parsed_table.rename_columns([x.root_fqn for x in features.values()])
            if len(parsed_table) > 0:
                yield parsed_table.combine_chunks().to_batches()[0]
            if len(parsed_table) == 0 and query_execution_parameters.yield_empty_batches:
                yield pa.RecordBatch.from_pydict({k: [] for k in parsed_table.schema.names}, parsed_table.schema)
            return

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_PGHOST_NAME, self.name, self.host),
                create_integration_variable(_PGPORT_NAME, self.name, self.port),
                create_integration_variable(_PGDATABASE_NAME, self.name, self.db),
                create_integration_variable(_PGUSER_NAME, self.name, self.user),
                create_integration_variable(_PGPASSWORD_NAME, self.name, self.password),
            ]
            if v is not None
        }


def _reformat_compiled_query_string_for_csv_output(
    stmt: str,
    *,
    escape_char: str,
    named_params: Mapping[str, Any],
) -> str:
    """
    - Replaces named parameters like `:foo` with the python3-style `{foo}` which is what `psycopg2.sql.SQL.format`
    expects to see in its inputs.

    - Wraps the provided `stmt` in a `COPY _ to STDOUT` declaration to format it as a CSV.
    """
    # Forcing quote so the unquoted empty string will represent null values

    # Convert the param style to python3-style {}, which is what psycopg2.sql.SQL.format expects

    original_stmt = stmt
    import sqlglot
    import sqlglot.expressions

    try:
        stmt_expression = sqlglot.parse_one(stmt, read="postgres")
        for placeholder in list(stmt_expression.find_all(sqlglot.expressions.Placeholder)):
            if isinstance(placeholder.this, str) and placeholder.this in named_params:
                # If this is a known placeholder like `:foo`, replace it verbatim with `{foo}`.
                # The use of the 'var' node here is incidental - it accepts any string as argument,
                # and sqlglot will reproduce that string verbatim as output.
                placeholder.replace(sqlglot.expressions.var("{" + placeholder.this + "}"))
        stmt = stmt_expression.sql(dialect="postgres")

        stmt = f"COPY ({stmt}) TO STDOUT (FORMAT CSV, HEADER true, FORCE_QUOTE *, ESCAPE '{escape_char}')"
    except Exception as e:
        raise ValueError(
            f"Failed to parse and convert postgres output {repr(original_stmt)} with named parameters {repr(list(named_params.keys()))} into valid psycopg2-compatible CSV OUTPUT statement: {e}"
        ) from e

    return stmt


def _get_df_with_parsed_dt_column(
    pl_table: pl.DataFrame, col_name: str, parser_formats: Sequence[str], target_dtype: PolarsTemporalType
):
    import polars as pl

    use_new_parsing = parse(pl.__version__) >= parse("0.18.4")
    num_nulls_original = pl_table[col_name].null_count()
    original_dt_clone: pl.Series = pl_table[col_name].clone()

    def parsing_complete(table: pl.DataFrame) -> bool:
        if table[col_name].dtype != target_dtype:
            # Never parsed
            return False
        return table[col_name].null_count() == num_nulls_original

    parser_formats_deque: Deque[Union[str, None]] = collections.deque(parser_formats)
    # When `format=None`, polars will raise a `ComputeError` if it
    # fails parsing, even if `strict=False`. Then, we will fallback
    # to the arrow CSV reader, which will raise a useful error.
    # TODO: Don't depend on pyarrow to raise pyarrow.ArrowInvalid,
    #       instead raise a custom parser error and handle that in
    #       the engine.
    parser_formats_deque.append(None)

    # strptime always outputs a datetime in UTC, so we do the
    # parsing in UTC first, and convert to target timezone later.
    parsing_dtype = target_dtype
    original_target_tz = None
    if target_dtype == pl.Datetime:
        target_dtype = cast(pl.Datetime, target_dtype)
        original_target_tz = target_dtype.time_zone
        if use_new_parsing:
            assert target_dtype.time_unit is not None, f"polars dtype {target_dtype} incorrectly set with no time unit."
            parsing_dtype = pl.Datetime(time_zone="UTC", time_unit=target_dtype.time_unit)
        else:
            target_dtype = pl.Datetime
            parsing_dtype = target_dtype

    # You might think this strptime call can be included in the for
    # loop below, but this is intentionally left out of the for loop
    # because we want to avoid calling `pl.coalesce` for the first pass.
    # Otherwise, the result of the coalescence will be a column of
    # `pl.Utf8` instead of the target date or datetime type.
    pl_table = pl_table.with_columns(
        original_dt_clone.str.strptime(
            cast("PolarsTemporalType", parsing_dtype),
            parser_formats_deque.popleft(),
            strict=False,
        ).alias(col_name)
    )

    # Incrementally coalesce so that we don't have to do all
    # passes of parsing - only as many as needed.
    while len(parser_formats_deque) > 0 and not parsing_complete(pl_table):
        fmt = parser_formats_deque.popleft()
        formatted_col = original_dt_clone.str.strptime(
            cast("PolarsTemporalType", parsing_dtype),
            fmt,
            strict=False,
        ).alias(col_name)
        coalesced_col = pl.coalesce(
            pl.col(col_name),
            formatted_col,
        )
        pl_table = pl_table.with_columns(coalesced_col)

    if target_dtype == pl.Datetime and use_new_parsing:
        # Convert to target timezone
        if original_target_tz is None:
            pl_table = pl_table.with_columns(pl.col(col_name).dt.replace_time_zone(None))
        else:
            pl_table = pl_table.with_columns(pl.col(col_name).dt.convert_time_zone(original_target_tz))

    # We would have raised if there was an error parsing with `format=None`.
    assert parsing_complete(pl_table)

    return pl_table



from typing import Any, Mapping, Optional, Tuple


class Validator:
    @staticmethod
    def string(value: Any, name: str, *, one_of: Optional[Tuple[str, ...]] = None) -> str:
        if value is None:
            raise ValueError(f"Expected a string, but found nothing for '{name}'")
        if not isinstance(value, str):
            raise ValueError(f"Expected a string, but found '{value}' for '{name}'")
        if one_of is not None and value not in one_of:
            raise ValueError(f"Expected '{name}' to be one of {one_of}, got '{value}'")
        return value

    @staticmethod
    def optional_string(value: Any, name: str, *, one_of: Optional[Tuple[str, ...]] = None) -> Optional[str]:
        return None if value is None else Validator.string(value=value, name=name, one_of=one_of)

    @staticmethod
    def dict_with_str_keys(value: Any, name: str) -> Mapping[str, Any]:
        if not isinstance(value, dict):
            raise ValueError(f"Expected dict for '{name}', got '{value}'")
        for k in value.keys():
            if not isinstance(k, str):
                raise ValueError(f"Expected string key for '{name}', got '{k}'")
        return value

    @staticmethod
    def dict_with_str_keys_or_none(value: Any, name: str) -> Optional[Mapping[str, Any]]:
        return None if value is None else Validator.dict_with_str_keys(value, name)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/flag.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1a\x63halk/server/v1/flag.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto"<\n\x10\x46\x65\x61tureFlagValue\x12\x12\n\x04\x66lag\x18\x01 \x01(\tR\x04\x66lag\x12\x14\n\x05value\x18\x02 \x01(\x08R\x05value"\x18\n\x16GetFeatureFlagsRequest"R\n\x17GetFeatureFlagsResponse\x12\x37\n\x05\x66lags\x18\x01 \x03(\x0b\x32!.chalk.server.v1.FeatureFlagValueR\x05\x66lags"s\n\x15SetFeatureFlagRequest\x12\x12\n\x04\x66lag\x18\x01 \x01(\tR\x04\x66lag\x12\x14\n\x05value\x18\x02 \x01(\x08R\x05value\x12\x30\n\x05scope\x18\x03 \x01(\x0e\x32\x1a.chalk.server.v1.FlagScopeR\x05scope"\x18\n\x16SetFeatureFlagResponse*X\n\tFlagScope\x12\x1a\n\x16\x46LAG_SCOPE_UNSPECIFIED\x10\x00\x12\x13\n\x0f\x46LAG_SCOPE_TEAM\x10\x01\x12\x1a\n\x16\x46LAG_SCOPE_ENVIRONMENT\x10\x02\x32\xed\x01\n\x12\x46\x65\x61tureFlagService\x12l\n\x0fGetFeatureFlags\x12\'.chalk.server.v1.GetFeatureFlagsRequest\x1a(.chalk.server.v1.GetFeatureFlagsResponse"\x06\x90\x02\x01\x80}\x02\x12i\n\x0eSetFeatureFlag\x12&.chalk.server.v1.SetFeatureFlagRequest\x1a\'.chalk.server.v1.SetFeatureFlagResponse"\x06\x90\x02\x02\x80}\x1b\x42\x92\x01\n\x13\x63om.chalk.server.v1B\tFlagProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.flag_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\tFlagProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_FEATUREFLAGSERVICE"].methods_by_name["GetFeatureFlags"]._options = None
    _globals["_FEATUREFLAGSERVICE"].methods_by_name["GetFeatureFlags"]._serialized_options = b"\220\002\001\200}\002"
    _globals["_FEATUREFLAGSERVICE"].methods_by_name["SetFeatureFlag"]._options = None
    _globals["_FEATUREFLAGSERVICE"].methods_by_name["SetFeatureFlag"]._serialized_options = b"\220\002\002\200}\033"
    _globals["_FLAGSCOPE"]._serialized_start = 395
    _globals["_FLAGSCOPE"]._serialized_end = 483
    _globals["_FEATUREFLAGVALUE"]._serialized_start = 80
    _globals["_FEATUREFLAGVALUE"]._serialized_end = 140
    _globals["_GETFEATUREFLAGSREQUEST"]._serialized_start = 142
    _globals["_GETFEATUREFLAGSREQUEST"]._serialized_end = 166
    _globals["_GETFEATUREFLAGSRESPONSE"]._serialized_start = 168
    _globals["_GETFEATUREFLAGSRESPONSE"]._serialized_end = 250
    _globals["_SETFEATUREFLAGREQUEST"]._serialized_start = 252
    _globals["_SETFEATUREFLAGREQUEST"]._serialized_end = 367
    _globals["_SETFEATUREFLAGRESPONSE"]._serialized_start = 369
    _globals["_SETFEATUREFLAGRESPONSE"]._serialized_end = 393
    _globals["_FEATUREFLAGSERVICE"]._serialized_start = 486
    _globals["_FEATUREFLAGSERVICE"]._serialized_end = 723
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import asyncio
import enum
import io
import os
import pathlib
import shutil
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from typing import (
    IO,
    TYPE_CHECKING,
    AsyncIterator,
    BinaryIO,
    Callable,
    ClassVar,
    Iterable,
    Mapping,
    Optional,
    Protocol,
    cast,
    overload,
)

import pyarrow.parquet as pq
from tenacity import RetryCallState, Retrying, stop_after_attempt, wait_fixed
from typing_extensions import assert_never, override

from chalk.utils.async_helpers import to_async_iterable
from chalk.utils.collections import unwrap_optional
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.tracing import safe_distribution, safe_incr, safe_trace

if TYPE_CHECKING:
    import google.cloud.storage
    import pyarrow as pa
    from fsspec import AbstractFileSystem
    from mypy_boto3_s3.client import S3Client


_logger = get_logger(__name__)


class _IOWithLen(io.BytesIO):
    def __len__(self):
        return len(self.getvalue())


class SignedUrlMode(enum.Enum):
    """Download or upload mode when requesting a signed URL"""

    DOWNLOAD = enum.auto()
    UPLOAD = enum.auto()


class StorageClient(Protocol):
    """Interface for a gcs / s3 client"""

    ...
    protocol: ClassVar[str]
    """The protocol for this storage client"""

    bucket: str
    """The bucket for this storage client"""

    fs: AbstractFileSystem
    """The fsspec filesystem for this storage client, e.g. used for parquet file metadata reading"""

    def upload_object(
        self,
        filename: str,
        content_type: str,
        data: bytes | BinaryIO,
        metadata: Mapping[str, str] | None = None,
    ):
        """Upload ``data`` to the ``filename``"""
        ...

    async def async_upload_object(
        self,
        filename: str,
        content_type: str,
        data: bytes | BinaryIO,
        metadata: Mapping[str, str] | None = None,
    ):
        ...

    def get_uri(self, filename: str) -> str:
        """Return a URI for a filename"""
        return f"{self.protocol}://{self.bucket}/{filename}"

    def get_filename(self, uri: str):
        """Return a filename for a uri"""
        return uri.replace(f"{self.protocol}://{self.bucket}/", "")

    def get_directory(self, uri: str):
        """NB: pathlib changes protocol strings"""
        return os.path.dirname(uri)

    @overload
    def download_object(self, filename: str) -> bytes:
        """Download ``filename`` and return the contents as bytes"""
        ...

    @overload
    def download_object(self, filename: str, destination: BinaryIO | str) -> None:
        """Download ``bucket/filename`` to the destination, which can either be an io object or a filepath"""
        ...

    @overload
    async def async_download_object(self, filename: str) -> bytes:
        """Download ``filename`` and return the contents as bytes"""
        ...

    @overload
    async def async_download_object(self, filename: str, destination: BinaryIO | str) -> None:
        """Download ``bucket/filename`` to the destination, which can either be an io object or a filepath"""
        ...

    def sign_url(
        self,
        filename: str,
        expiration: datetime,
        mode: SignedUrlMode,
        response_disposition: str | None = None,
    ) -> str:
        """
        Generate a signed url for ``filename`` that expires at ``expiration``
        """
        ...

    def list_files(self, prefix: str, delimiter: Optional[str] = None) -> Iterable[str]:
        """
        Yield filenames from the bucket that begin with the given prefix.
        Please use 'async_list_files` when possible (in an async function)
        """
        ...

    def async_list_files(self, prefix: str, delimiter: Optional[str] = None) -> AsyncIterator[str]:
        ...

    def copy(self, source_filename: str, dest_filename: str) -> None:
        """Copy ``source_filename`` to dest_filename``, including all metadata"""
        ...

    def get_file_size(self, filename: str) -> int:
        """Get the file size, in bytes"""
        ...


def _log_upload_retry(state: RetryCallState) -> None:
    if state.outcome is not None:
        safe_incr(
            counter="chalk.engine.gcs_bulk_upload_output.retry",
            value=state.attempt_number,
            tags=[f"success:{not state.outcome.failed}"],
        )


class ParquetStorageClient:
    def __init__(self, fs: AbstractFileSystem, file_normalizer_fn: Callable[[str], str]):
        super().__init__()
        self.fs = fs
        self.file_normalizer_fn = file_normalizer_fn

    def _get_parquet_lazy(self, filename: str) -> pq.ParquetFile:
        normalized_filename = self.file_normalizer_fn(filename)
        with self.fs.open(normalized_filename, "rb") as f:
            return pq.ParquetFile(f)

    def get_num_rows_parquet(self, filename: str) -> int:
        parquet_file = self._get_parquet_lazy(filename)
        assert parquet_file.metadata is not None, "no parquet metadata found"
        return parquet_file.metadata.num_rows

    def get_schema_parquet(self, filename: str) -> pa.Schema:
        parquet_file = self._get_parquet_lazy(filename)
        return parquet_file.schema_arrow

    @classmethod
    def from_storage_client(cls, storage_client: StorageClient):
        return cls(fs=storage_client.fs, file_normalizer_fn=storage_client.get_uri)


class GCSStorageClient(StorageClient):
    protocol = "gs"

    def __init__(
        self,
        gcs_client: google.cloud.storage.Client,
        gcs_executor: ThreadPoolExecutor,
        bucket: str,
    ):
        super().__init__()
        try:
            import gcsfs
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")
        self.bucket = bucket
        self._gcs_client = gcs_client
        self._gcs_executor = gcs_executor
        self._bucket = self._gcs_client.bucket(bucket)
        self.fs = gcsfs.GCSFileSystem()

    @override
    def upload_object(
        self,
        filename: str,
        content_type: str,
        data: bytes | BinaryIO,
        metadata: Mapping[str, str] | None = None,
    ):
        if isinstance(data, bytes):
            data = io.BytesIO(data)
        assert isinstance(data, io.IOBase)
        data.seek(0, os.SEEK_END)
        num_bytes = data.tell()
        data.seek(0)
        file_uri = self.get_uri(filename)
        blob = self._bucket.blob(filename)
        blob.metadata = metadata
        _logger.debug(f"Uploading file '{file_uri}' of size {num_bytes} bytes")

        # Empirically this appears to fail occasionally, so let's just try it five times.
        # FIXME: Configure the retrying on the google storage client, instead of doing our own wrapper around it
        for attempt in Retrying(
            stop=stop_after_attempt(5),
            reraise=True,
            wait=wait_fixed(3),
            after=_log_upload_retry,
        ):
            # matters for retries
            data.seek(0)
            with attempt:
                with safe_trace(
                    "CloudStorageBulkUploadOutput.do_upload",
                    attributes={
                        "size_bytes": str(num_bytes),
                        "file_uri": file_uri,
                    },
                ):
                    # Not using the actual filename in the tag to avoid tag blowup
                    safe_distribution(
                        "chalk.engine.storage_client.files_uploaded",
                        1,
                        tags=[f"bucket:{self.bucket}", f"protocol:{self.protocol}"],
                    )
                    safe_distribution(
                        "chalk.engine.storage_client.bytes_uploaded",
                        num_bytes,
                        tags=[f"bucket:{self.bucket}", f"protocol:{self.protocol}"],
                    )
                    blob.upload_from_file(
                        data,
                        size=num_bytes,
                        content_type=content_type,
                        if_generation_match=blob.generation,
                    )

    @overload
    @override
    def download_object(self, filename: str) -> bytes:
        ...

    @overload
    @override
    def download_object(self, filename: str, destination: BinaryIO | str) -> None:
        ...

    @override
    def download_object(self, filename: str, destination: str | BinaryIO | None = None) -> bytes | None:
        blob = self._bucket.blob(filename)
        if destination is None:
            return blob.download_as_bytes()
        elif isinstance(destination, str):
            blob.download_to_filename(destination)
        else:
            blob.download_to_file(destination)

    async def async_upload_object(
        self,
        filename: str,
        content_type: str,
        data: bytes | BinaryIO,
        metadata: Mapping[str, str] | None = None,
    ):
        return await asyncio.get_running_loop().run_in_executor(
            self._gcs_executor,
            self.upload_object,
            filename,
            content_type,
            data,
            metadata,
        )

    @overload
    @override
    async def async_download_object(self, filename: str) -> bytes:
        ...

    @overload
    @override
    async def async_download_object(self, filename: str, destination: BinaryIO | str) -> None:
        ...

    @override
    async def async_download_object(self, filename: str, destination: str | BinaryIO | None = None) -> bytes | None:
        if destination is None:
            return cast(
                None,
                await asyncio.get_running_loop().run_in_executor(self._gcs_executor, self.download_object, filename),
            )
        else:
            return await asyncio.get_running_loop().run_in_executor(
                self._gcs_executor, self.download_object, filename, destination
            )

    def sign_url(
        self,
        filename: str,
        expiration: datetime,
        mode: SignedUrlMode,
        response_disposition: str | None = None,
    ) -> str:
        blob = self._bucket.blob(filename)
        if mode == SignedUrlMode.DOWNLOAD:
            method = "GET"
        elif mode == SignedUrlMode.UPLOAD:
            method = "PUT"
        else:
            assert_never(mode)
        try:
            return blob.generate_signed_url(
                expiration=expiration,
                method=method,
                version="v4",
                response_disposition=response_disposition,
            )
        except Exception as e:
            _logger.error(
                f"Failed to create signed url for '{filename}'; falling back to storage client URL",
                exc_info=e,
            )

        return self.get_uri(filename)

    def list_files(self, prefix: str, delimiter: Optional[str] = None) -> Iterable[str]:
        blobs = self._gcs_client.list_blobs(self.bucket, prefix=prefix, delimiter=delimiter)
        for blob in blobs:
            yield blob.name

    async def async_list_files(self, prefix: str, delimiter: Optional[str] = None) -> AsyncIterator[str]:
        iterable = await asyncio.get_running_loop().run_in_executor(
            self._gcs_executor, self.list_files, prefix, delimiter
        )
        async for filename in to_async_iterable(iterable, self._gcs_executor):
            yield filename

    def copy(self, source_filename: str, dest_filename: str) -> None:
        blob = self._bucket.blob(source_filename)
        self._bucket.copy_blob(blob, destination_bucket=self._bucket, new_name=dest_filename)

    def get_file_size(self, filename: str) -> int:
        blob = self._bucket.get_blob(filename)
        assert blob is not None
        size = blob.size
        assert size is not None
        return size


class S3StorageClient(StorageClient):
    protocol = "s3"

    def __init__(
        self,
        bucket: str,
        s3_client: S3Client,
        executor: ThreadPoolExecutor,
        aws_region: str | None = None,
    ) -> None:
        super().__init__()
        try:
            import s3fs
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")
        self.bucket = bucket
        self._s3_client = s3_client
        self._executor = executor
        region = aws_region or os.environ.get("AWS_DEFAULT_REGION")
        _logger.info(f"AWS storage client using {region=}")
        self.fs = s3fs.S3FileSystem(
            client_kwargs={"region_name": region},
        )

    def upload_object(
        self,
        filename: str,
        content_type: str,
        data: bytes | BinaryIO,
        metadata: Mapping[str, str] | None = None,
    ):
        if not isinstance(data, bytes):
            data.seek(0)
            data = data.read()
        assert isinstance(data, bytes)
        wrapped_data = _IOWithLen(
            data
        )  # boto3 wants to call len() the io, manually wrapping it in a class that supports that
        if metadata is None:
            metadata = {}
        # Not using the actual filename in the tag to avoid tag blowup
        safe_distribution(
            "chalk.engine.storage_client.files_uploaded",
            1,
            tags=[f"bucket:{self.bucket}", f"protocol:{self.protocol}"],
        )
        safe_distribution(
            "chalk.engine.storage_client.bytes_uploaded",
            len(data),
            tags=[f"bucket:{self.bucket}", f"protocol:{self.protocol}"],
        )

        if len(wrapped_data) <= 5 * 1024 * 1024 - 1000:  # 5MB is the limit for a single PUT request
            self._s3_client.put_object(
                Bucket=self.bucket,
                Key=filename,
                Body=wrapped_data,
                ContentType=content_type,
                Metadata=metadata,
            )
        else:
            # Multipart upload
            multipart_upload = self._s3_client.create_multipart_upload(
                Bucket=self.bucket,
                Key=filename,
                ContentType=content_type,
                Metadata=metadata,
            )
            upload_id = multipart_upload["UploadId"]
            part_size = 100 * 1024 * 1024  # 100MB
            parts = []

            for i in range(0, len(data), part_size):
                part_number = i // part_size + 1
                part_data = data[i : i + part_size]
                part = self._s3_client.upload_part(
                    Bucket=self.bucket,
                    Key=filename,
                    PartNumber=part_number,
                    UploadId=upload_id,
                    Body=part_data,
                )
                parts.append(
                    {
                        "PartNumber": part_number,
                        "ETag": part["ETag"],
                    }
                )

            self._s3_client.complete_multipart_upload(
                Bucket=self.bucket,
                Key=filename,
                UploadId=upload_id,
                MultipartUpload={"Parts": parts},
            )

    @overload
    @override
    def download_object(self, filename: str) -> bytes:
        ...

    @overload
    @override
    def download_object(self, filename: str, destination: BinaryIO | str) -> None:
        ...

    def _download_to_file(self, filename: str, output_filename: str):
        self._s3_client.download_file(
            Bucket=self.bucket,
            Key=filename,
            Filename=output_filename,
        )

    def _download_to_fileobj(self, filename: str, output_file: IO):
        self._s3_client.download_fileobj(
            Bucket=self.bucket,
            Key=filename,
            Fileobj=output_file,
        )

    @override
    def download_object(self, filename: str, destination: str | BinaryIO | None = None) -> bytes | None:
        if isinstance(destination, str):
            self._download_to_file(filename, destination)
            return
        if destination is None:
            destination = io.BytesIO()
            self._download_to_fileobj(filename, destination)
            destination.seek(0)
            return destination.read()
        else:
            self._download_to_fileobj(filename, destination)

    async def async_upload_object(
        self,
        filename: str,
        content_type: str,
        data: bytes | BinaryIO,
        metadata: Mapping[str, str] | None = None,
    ):
        return await asyncio.get_running_loop().run_in_executor(
            self._executor, self.upload_object, filename, content_type, data, metadata
        )

    @overload
    @override
    async def async_download_object(self, filename: str) -> bytes:
        ...

    @overload
    @override
    async def async_download_object(self, filename: str, destination: BinaryIO | str) -> None:
        ...

    @override
    async def async_download_object(self, filename: str, destination: str | BinaryIO | None = None) -> bytes | None:
        if destination is None:
            return cast(
                None,
                await asyncio.get_running_loop().run_in_executor(self._executor, self.download_object, filename),
            )
        else:
            return await asyncio.get_running_loop().run_in_executor(
                self._executor, self.download_object, filename, destination
            )

    def sign_url(
        self,
        filename: str,
        expiration: datetime,
        mode: SignedUrlMode,
        response_disposition: str | None = None,
    ) -> str:
        expiration_secs = (expiration - datetime.now(timezone.utc)).total_seconds()
        if expiration_secs <= 0:
            raise ValueError("Expiration time is in the past")
        params = {"Bucket": self.bucket, "Key": filename}
        if response_disposition is not None:
            params["ResponseContentDisposition"] = response_disposition
        if mode == SignedUrlMode.DOWNLOAD:
            client_method = "get_object"
        elif mode == SignedUrlMode.UPLOAD:
            client_method = "put_object"
        else:
            assert_never(mode)
        return self._s3_client.generate_presigned_url(
            ClientMethod=client_method,
            Params=params,
            ExpiresIn=int(expiration_secs),
        )

    def list_files(self, prefix: str, delimiter: Optional[str] = None) -> Iterable[str]:
        try:
            if delimiter is None:
                continuation_token = None
                while True:
                    if continuation_token is None:
                        resp = self._s3_client.list_objects_v2(
                            Bucket=self.bucket,
                            Prefix=prefix,
                        )
                    else:
                        resp = self._s3_client.list_objects_v2(
                            Bucket=self.bucket,
                            Prefix=prefix,
                            ContinuationToken=continuation_token,
                        )
                    # If no keys returned, the server omits 'Contents'
                    for row in resp.get("Contents", []):
                        key = row.get("Key")
                        assert key is not None, "all objects must have a key"
                        yield key
                    if not resp["IsTruncated"]:
                        return
                    continuation_token = resp["NextContinuationToken"]
            else:
                resp = self._s3_client.list_objects_v2(
                    Bucket=self.bucket,
                    Prefix=prefix,
                    Delimiter=delimiter,
                )
                if "CommonPrefixes" not in resp:
                    return  # s3 doesn't include the key if empty
                for result in resp["CommonPrefixes"]:
                    yield unwrap_optional(result.get("Prefix"))
        except Exception:
            _logger.error(f"Got exception while listing files for {prefix=}", exc_info=True)
            raise

    async def async_list_files(self, prefix: str, delimiter: Optional[str] = None):
        iterable = await asyncio.get_running_loop().run_in_executor(self._executor, self.list_files, prefix, delimiter)
        async for filename in to_async_iterable(iterable, self._executor):
            yield filename

    def copy(self, source_filename: str, dest_filename: str) -> None:
        self._s3_client.copy(
            CopySource={
                "Bucket": self.bucket,
                "Key": source_filename,
            },
            Bucket=self.bucket,
            Key=dest_filename,
        )

    def get_file_size(self, filename: str) -> int:
        obj = self._s3_client.get_object(
            Bucket=self.bucket,
            Key=filename,
        )
        return obj["ContentLength"]


class LocalStorageClient(StorageClient):
    protocol = "file"

    def __init__(self, folder: str):
        super().__init__()
        try:
            import fsspec
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")
        self.bucket = ""
        self._folder = pathlib.Path(os.path.abspath(folder))
        self.fs = fsspec.filesystem("file")

    def _normalize_path(self, prefix: str):
        if prefix.startswith("file://"):
            return pathlib.Path(prefix[len("file://") :])
        if prefix.lstrip("/").startswith(str(self._folder).lstrip("/")):
            # It's possible that the prefixes were derived from a uri, in which case the full path is being passed in
            if not prefix.startswith("/"):
                prefix = "/" + prefix
            return pathlib.Path(prefix)
        return self._folder / prefix

    def list_files(self, prefix: str, delimiter: str | None = None) -> Iterable[str]:
        if delimiter is not None:
            raise ValueError("delimiter is not supported for the local storage client")
        joined = self._normalize_path(prefix)
        filename_prefix = ""
        parent = joined
        if not joined.exists():
            parent = joined.parent
            filename_prefix = joined.name
        if not parent.exists():
            return

        # Joined could be a folder, file, or folder with filename prefix
        if joined.is_file():
            yield prefix
            return

        for root, _, files in os.walk(parent):
            for name in files:
                if os.path.relpath(os.path.join(root, name), parent).startswith(filename_prefix):
                    full_path = os.path.join(root, name)
                    rel_path = os.path.relpath(full_path, self._folder)
                    yield rel_path

    def sign_url(
        self,
        filename: str,
        expiration: datetime,
        mode: SignedUrlMode,
        response_disposition: str | None = None,
    ) -> str:
        return "file://" + str(self._normalize_path(filename))

    def upload_object(
        self,
        filename: str,
        content_type: str,
        data: bytes | BinaryIO,
        metadata: Mapping[str, str] | None = None,
    ):
        """Upload ``data`` to the ``filename``"""
        full_path = self._normalize_path(filename)
        full_path.parent.mkdir(parents=True, exist_ok=True)
        with open(full_path, "wb") as f:
            if not isinstance(data, bytes):
                data = data.read()
            f.write(data)

    async def async_upload_object(
        self,
        filename: str,
        content_type: str,
        data: bytes | BinaryIO,
        metadata: Mapping[str, str] | None = None,
    ):
        return self.upload_object(filename, content_type, data, metadata)

    def get_uri(self, filename: str):
        """Return a URI for a filename"""
        return f"{self.protocol}://" + str(self._normalize_path(filename))

    @overload
    @override
    def download_object(self, filename: str) -> bytes:
        ...

    @overload
    @override
    def download_object(self, filename: str, destination: BinaryIO | str) -> None:
        ...

    def download_object(self, filename: str, destination: BinaryIO | str | None = None) -> bytes | None:
        """Download ``bucket/filename`` to the destination, which can either be an io object or a filepath"""
        path = self._normalize_path(filename)
        if destination is None:
            with open(path, "rb") as f:
                return f.read()
        if isinstance(destination, str):
            shutil.copy2(path, destination)
        else:
            with open(path, "rb") as f:
                destination.write(f.read())

    @overload
    @override
    async def async_download_object(self, filename: str) -> bytes:
        ...

    @overload
    @override
    async def async_download_object(self, filename: str, destination: BinaryIO | str) -> None:
        ...

    async def async_download_object(self, filename: str, destination: BinaryIO | str | None = None):
        """Download ``filename`` and return the contents as bytes"""
        if destination is None:
            return self.download_object(filename)
        else:
            return self.download_object(filename, destination)

    async def async_list_files(self, prefix: str, delimiter: str | None = None) -> AsyncIterator[str]:
        for x in self.list_files(prefix, delimiter):
            yield x

    def copy(self, source_filename: str, dest_filename: str):
        shutil.copy2(self._normalize_path(source_filename), self._normalize_path(dest_filename))

    def get_file_size(self, filename: str):
        return os.stat(self._normalize_path(filename)).st_size



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf import field_mask_pb2 as _field_mask_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class PromptTemplate(_message.Message):
    __slots__ = ("type", "message")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    type: str
    message: str
    def __init__(self, type: _Optional[str] = ..., message: _Optional[str] = ...) -> None: ...

class ProviderConfig(_message.Message):
    __slots__ = ("provider", "model_name", "parameters", "structured_output")
    class ParametersEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _struct_pb2.Value
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...
        ) -> None: ...

    PROVIDER_FIELD_NUMBER: _ClassVar[int]
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    STRUCTURED_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    provider: str
    model_name: str
    parameters: _containers.MessageMap[str, _struct_pb2.Value]
    structured_output: str
    def __init__(
        self,
        provider: _Optional[str] = ...,
        model_name: _Optional[str] = ...,
        parameters: _Optional[_Mapping[str, _struct_pb2.Value]] = ...,
        structured_output: _Optional[str] = ...,
    ) -> None: ...

class PromptVariant(_message.Message):
    __slots__ = (
        "id",
        "environment_id",
        "deployment_id",
        "named_prompt_id",
        "evaluation_run_id",
        "commit_hash",
        "templates",
        "variables",
        "provider_config",
        "variant_hash",
        "created_by",
        "created_at",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    NAMED_PROMPT_ID_FIELD_NUMBER: _ClassVar[int]
    EVALUATION_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    COMMIT_HASH_FIELD_NUMBER: _ClassVar[int]
    TEMPLATES_FIELD_NUMBER: _ClassVar[int]
    VARIABLES_FIELD_NUMBER: _ClassVar[int]
    PROVIDER_CONFIG_FIELD_NUMBER: _ClassVar[int]
    VARIANT_HASH_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    id: int
    environment_id: str
    deployment_id: str
    named_prompt_id: str
    evaluation_run_id: str
    commit_hash: str
    templates: _containers.RepeatedCompositeFieldContainer[PromptTemplate]
    variables: _containers.RepeatedScalarFieldContainer[str]
    provider_config: ProviderConfig
    variant_hash: str
    created_by: str
    created_at: _timestamp_pb2.Timestamp
    def __init__(
        self,
        id: _Optional[int] = ...,
        environment_id: _Optional[str] = ...,
        deployment_id: _Optional[str] = ...,
        named_prompt_id: _Optional[str] = ...,
        evaluation_run_id: _Optional[str] = ...,
        commit_hash: _Optional[str] = ...,
        templates: _Optional[_Iterable[_Union[PromptTemplate, _Mapping]]] = ...,
        variables: _Optional[_Iterable[str]] = ...,
        provider_config: _Optional[_Union[ProviderConfig, _Mapping]] = ...,
        variant_hash: _Optional[str] = ...,
        created_by: _Optional[str] = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class NamedPrompt(_message.Message):
    __slots__ = (
        "id",
        "environment_id",
        "deployment_id",
        "name",
        "description",
        "tags",
        "created_by",
        "created_at",
        "updated_at",
        "archived_at",
        "latest_prompt_variant",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    ARCHIVED_AT_FIELD_NUMBER: _ClassVar[int]
    LATEST_PROMPT_VARIANT_FIELD_NUMBER: _ClassVar[int]
    id: str
    environment_id: str
    deployment_id: str
    name: str
    description: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    created_by: str
    created_at: _timestamp_pb2.Timestamp
    updated_at: _timestamp_pb2.Timestamp
    archived_at: _timestamp_pb2.Timestamp
    latest_prompt_variant: PromptVariant
    def __init__(
        self,
        id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        deployment_id: _Optional[str] = ...,
        name: _Optional[str] = ...,
        description: _Optional[str] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        created_by: _Optional[str] = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        archived_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        latest_prompt_variant: _Optional[_Union[PromptVariant, _Mapping]] = ...,
    ) -> None: ...

class PromptEvaluationRun(_message.Message):
    __slots__ = (
        "id",
        "environment_id",
        "deployment_id",
        "branch_name",
        "dataset_id",
        "dataset_revision_id",
        "reference_output",
        "evaluators",
        "offline_query_ids",
        "related_named_prompt_ids",
        "related_evaluation_ids",
        "aggregate_metrics",
        "meta_data",
        "created_by",
        "created_at",
        "dataset_name",
        "related_named_prompt_names",
        "prompt_variants",
    )
    class AggregateMetricsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _struct_pb2.Value
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...
        ) -> None: ...

    class MetaDataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _struct_pb2.Value
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...
        ) -> None: ...

    ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    BRANCH_NAME_FIELD_NUMBER: _ClassVar[int]
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    DATASET_REVISION_ID_FIELD_NUMBER: _ClassVar[int]
    REFERENCE_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    EVALUATORS_FIELD_NUMBER: _ClassVar[int]
    OFFLINE_QUERY_IDS_FIELD_NUMBER: _ClassVar[int]
    RELATED_NAMED_PROMPT_IDS_FIELD_NUMBER: _ClassVar[int]
    RELATED_EVALUATION_IDS_FIELD_NUMBER: _ClassVar[int]
    AGGREGATE_METRICS_FIELD_NUMBER: _ClassVar[int]
    META_DATA_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    DATASET_NAME_FIELD_NUMBER: _ClassVar[int]
    RELATED_NAMED_PROMPT_NAMES_FIELD_NUMBER: _ClassVar[int]
    PROMPT_VARIANTS_FIELD_NUMBER: _ClassVar[int]
    id: str
    environment_id: str
    deployment_id: str
    branch_name: str
    dataset_id: str
    dataset_revision_id: str
    reference_output: str
    evaluators: _containers.RepeatedScalarFieldContainer[str]
    offline_query_ids: _containers.RepeatedScalarFieldContainer[str]
    related_named_prompt_ids: _containers.RepeatedScalarFieldContainer[str]
    related_evaluation_ids: _containers.RepeatedScalarFieldContainer[str]
    aggregate_metrics: _containers.MessageMap[str, _struct_pb2.Value]
    meta_data: _containers.MessageMap[str, _struct_pb2.Value]
    created_by: str
    created_at: _timestamp_pb2.Timestamp
    dataset_name: str
    related_named_prompt_names: _containers.RepeatedScalarFieldContainer[str]
    prompt_variants: _containers.RepeatedCompositeFieldContainer[PromptVariant]
    def __init__(
        self,
        id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        deployment_id: _Optional[str] = ...,
        branch_name: _Optional[str] = ...,
        dataset_id: _Optional[str] = ...,
        dataset_revision_id: _Optional[str] = ...,
        reference_output: _Optional[str] = ...,
        evaluators: _Optional[_Iterable[str]] = ...,
        offline_query_ids: _Optional[_Iterable[str]] = ...,
        related_named_prompt_ids: _Optional[_Iterable[str]] = ...,
        related_evaluation_ids: _Optional[_Iterable[str]] = ...,
        aggregate_metrics: _Optional[_Mapping[str, _struct_pb2.Value]] = ...,
        meta_data: _Optional[_Mapping[str, _struct_pb2.Value]] = ...,
        created_by: _Optional[str] = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        dataset_name: _Optional[str] = ...,
        related_named_prompt_names: _Optional[_Iterable[str]] = ...,
        prompt_variants: _Optional[_Iterable[_Union[PromptVariant, _Mapping]]] = ...,
    ) -> None: ...

class ListNamedPromptsRequest(_message.Message):
    __slots__ = ("cursor", "limit")
    CURSOR_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    cursor: str
    limit: int
    def __init__(self, cursor: _Optional[str] = ..., limit: _Optional[int] = ...) -> None: ...

class ListNamedPromptsResponse(_message.Message):
    __slots__ = ("named_prompts", "next_cursor")
    NAMED_PROMPTS_FIELD_NUMBER: _ClassVar[int]
    NEXT_CURSOR_FIELD_NUMBER: _ClassVar[int]
    named_prompts: _containers.RepeatedCompositeFieldContainer[NamedPrompt]
    next_cursor: str
    def __init__(
        self,
        named_prompts: _Optional[_Iterable[_Union[NamedPrompt, _Mapping]]] = ...,
        next_cursor: _Optional[str] = ...,
    ) -> None: ...

class GetNamedPromptRequest(_message.Message):
    __slots__ = ("named_prompt_id",)
    NAMED_PROMPT_ID_FIELD_NUMBER: _ClassVar[int]
    named_prompt_id: str
    def __init__(self, named_prompt_id: _Optional[str] = ...) -> None: ...

class GetNamedPromptResponse(_message.Message):
    __slots__ = ("named_prompt",)
    NAMED_PROMPT_FIELD_NUMBER: _ClassVar[int]
    named_prompt: NamedPrompt
    def __init__(self, named_prompt: _Optional[_Union[NamedPrompt, _Mapping]] = ...) -> None: ...

class PromptVariantOperation(_message.Message):
    __slots__ = ("templates", "provider_config")
    TEMPLATES_FIELD_NUMBER: _ClassVar[int]
    PROVIDER_CONFIG_FIELD_NUMBER: _ClassVar[int]
    templates: _containers.RepeatedCompositeFieldContainer[PromptTemplate]
    provider_config: ProviderConfig
    def __init__(
        self,
        templates: _Optional[_Iterable[_Union[PromptTemplate, _Mapping]]] = ...,
        provider_config: _Optional[_Union[ProviderConfig, _Mapping]] = ...,
    ) -> None: ...

class CreateNamedPromptRequest(_message.Message):
    __slots__ = ("name", "description", "tags", "prompt_variant")
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    PROMPT_VARIANT_FIELD_NUMBER: _ClassVar[int]
    name: str
    description: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    prompt_variant: PromptVariantOperation
    def __init__(
        self,
        name: _Optional[str] = ...,
        description: _Optional[str] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        prompt_variant: _Optional[_Union[PromptVariantOperation, _Mapping]] = ...,
    ) -> None: ...

class CreateNamedPromptResponse(_message.Message):
    __slots__ = ("named_prompt",)
    NAMED_PROMPT_FIELD_NUMBER: _ClassVar[int]
    named_prompt: NamedPrompt
    def __init__(self, named_prompt: _Optional[_Union[NamedPrompt, _Mapping]] = ...) -> None: ...

class UpdateNamedPromptOperation(_message.Message):
    __slots__ = ("name", "description", "tags", "prompt_variant", "archived_at")
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    PROMPT_VARIANT_FIELD_NUMBER: _ClassVar[int]
    ARCHIVED_AT_FIELD_NUMBER: _ClassVar[int]
    name: str
    description: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    prompt_variant: PromptVariantOperation
    archived_at: _timestamp_pb2.Timestamp
    def __init__(
        self,
        name: _Optional[str] = ...,
        description: _Optional[str] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        prompt_variant: _Optional[_Union[PromptVariantOperation, _Mapping]] = ...,
        archived_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class UpdateNamedPromptRequest(_message.Message):
    __slots__ = ("named_prompt_id", "update", "update_mask")
    NAMED_PROMPT_ID_FIELD_NUMBER: _ClassVar[int]
    UPDATE_FIELD_NUMBER: _ClassVar[int]
    UPDATE_MASK_FIELD_NUMBER: _ClassVar[int]
    named_prompt_id: str
    update: UpdateNamedPromptOperation
    update_mask: _field_mask_pb2.FieldMask
    def __init__(
        self,
        named_prompt_id: _Optional[str] = ...,
        update: _Optional[_Union[UpdateNamedPromptOperation, _Mapping]] = ...,
        update_mask: _Optional[_Union[_field_mask_pb2.FieldMask, _Mapping]] = ...,
    ) -> None: ...

class UpdateNamedPromptResponse(_message.Message):
    __slots__ = ("named_prompt",)
    NAMED_PROMPT_FIELD_NUMBER: _ClassVar[int]
    named_prompt: NamedPrompt
    def __init__(self, named_prompt: _Optional[_Union[NamedPrompt, _Mapping]] = ...) -> None: ...

class ListPromptVariantsRequest(_message.Message):
    __slots__ = ("named_prompt_id", "evaluation_run_id", "cursor", "limit")
    NAMED_PROMPT_ID_FIELD_NUMBER: _ClassVar[int]
    EVALUATION_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    CURSOR_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    named_prompt_id: str
    evaluation_run_id: str
    cursor: str
    limit: int
    def __init__(
        self,
        named_prompt_id: _Optional[str] = ...,
        evaluation_run_id: _Optional[str] = ...,
        cursor: _Optional[str] = ...,
        limit: _Optional[int] = ...,
    ) -> None: ...

class ListPromptVariantsResponse(_message.Message):
    __slots__ = ("prompt_variants", "next_cursor")
    PROMPT_VARIANTS_FIELD_NUMBER: _ClassVar[int]
    NEXT_CURSOR_FIELD_NUMBER: _ClassVar[int]
    prompt_variants: _containers.RepeatedCompositeFieldContainer[PromptVariant]
    next_cursor: str
    def __init__(
        self,
        prompt_variants: _Optional[_Iterable[_Union[PromptVariant, _Mapping]]] = ...,
        next_cursor: _Optional[str] = ...,
    ) -> None: ...

class ListPromptEvaluationRunsRequest(_message.Message):
    __slots__ = ("cursor", "limit", "named_prompt_id")
    CURSOR_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    NAMED_PROMPT_ID_FIELD_NUMBER: _ClassVar[int]
    cursor: str
    limit: int
    named_prompt_id: str
    def __init__(
        self, cursor: _Optional[str] = ..., limit: _Optional[int] = ..., named_prompt_id: _Optional[str] = ...
    ) -> None: ...

class ListPromptEvaluationRunsResponse(_message.Message):
    __slots__ = ("evaluation_runs", "next_cursor")
    EVALUATION_RUNS_FIELD_NUMBER: _ClassVar[int]
    NEXT_CURSOR_FIELD_NUMBER: _ClassVar[int]
    evaluation_runs: _containers.RepeatedCompositeFieldContainer[PromptEvaluationRun]
    next_cursor: str
    def __init__(
        self,
        evaluation_runs: _Optional[_Iterable[_Union[PromptEvaluationRun, _Mapping]]] = ...,
        next_cursor: _Optional[str] = ...,
    ) -> None: ...

class GetPromptEvaluationRunRequest(_message.Message):
    __slots__ = ("evaluation_run_id", "prefill_options")
    class PrefillOptions(_message.Message):
        __slots__ = ("named_prompt_id", "evaluation_run_id")
        NAMED_PROMPT_ID_FIELD_NUMBER: _ClassVar[int]
        EVALUATION_RUN_ID_FIELD_NUMBER: _ClassVar[int]
        named_prompt_id: str
        evaluation_run_id: str
        def __init__(self, named_prompt_id: _Optional[str] = ..., evaluation_run_id: _Optional[str] = ...) -> None: ...

    EVALUATION_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    PREFILL_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    evaluation_run_id: str
    prefill_options: GetPromptEvaluationRunRequest.PrefillOptions
    def __init__(
        self,
        evaluation_run_id: _Optional[str] = ...,
        prefill_options: _Optional[_Union[GetPromptEvaluationRunRequest.PrefillOptions, _Mapping]] = ...,
    ) -> None: ...

class GetPromptEvaluationRunResponse(_message.Message):
    __slots__ = ("evaluation_run",)
    EVALUATION_RUN_FIELD_NUMBER: _ClassVar[int]
    evaluation_run: PromptEvaluationRun
    def __init__(self, evaluation_run: _Optional[_Union[PromptEvaluationRun, _Mapping]] = ...) -> None: ...

class CreatePromptEvaluationRunRequest(_message.Message):
    __slots__ = (
        "branch_name",
        "dataset_id",
        "dataset_revision_id",
        "reference_output",
        "evaluators",
        "related_named_prompt_ids",
        "related_evaluation_ids",
        "meta_data",
        "prompt_variants",
    )
    class MetaDataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _struct_pb2.Value
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...
        ) -> None: ...

    BRANCH_NAME_FIELD_NUMBER: _ClassVar[int]
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    DATASET_REVISION_ID_FIELD_NUMBER: _ClassVar[int]
    REFERENCE_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    EVALUATORS_FIELD_NUMBER: _ClassVar[int]
    RELATED_NAMED_PROMPT_IDS_FIELD_NUMBER: _ClassVar[int]
    RELATED_EVALUATION_IDS_FIELD_NUMBER: _ClassVar[int]
    META_DATA_FIELD_NUMBER: _ClassVar[int]
    PROMPT_VARIANTS_FIELD_NUMBER: _ClassVar[int]
    branch_name: str
    dataset_id: str
    dataset_revision_id: str
    reference_output: str
    evaluators: _containers.RepeatedScalarFieldContainer[str]
    related_named_prompt_ids: _containers.RepeatedScalarFieldContainer[str]
    related_evaluation_ids: _containers.RepeatedScalarFieldContainer[str]
    meta_data: _containers.MessageMap[str, _struct_pb2.Value]
    prompt_variants: _containers.RepeatedCompositeFieldContainer[PromptVariantOperation]
    def __init__(
        self,
        branch_name: _Optional[str] = ...,
        dataset_id: _Optional[str] = ...,
        dataset_revision_id: _Optional[str] = ...,
        reference_output: _Optional[str] = ...,
        evaluators: _Optional[_Iterable[str]] = ...,
        related_named_prompt_ids: _Optional[_Iterable[str]] = ...,
        related_evaluation_ids: _Optional[_Iterable[str]] = ...,
        meta_data: _Optional[_Mapping[str, _struct_pb2.Value]] = ...,
        prompt_variants: _Optional[_Iterable[_Union[PromptVariantOperation, _Mapping]]] = ...,
    ) -> None: ...

class CreatePromptEvaluationRunResponse(_message.Message):
    __slots__ = ("evaluation_run",)
    EVALUATION_RUN_FIELD_NUMBER: _ClassVar[int]
    evaluation_run: PromptEvaluationRun
    def __init__(self, evaluation_run: _Optional[_Union[PromptEvaluationRun, _Mapping]] = ...) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.graph_pb2 import (
    GetCodegenFeaturesFromGraphRequest,
    GetCodegenFeaturesFromGraphResponse,
    GetFeatureSQLRequest,
    GetFeatureSQLResponse,
    GetFeaturesMetadataRequest,
    GetFeaturesMetadataResponse,
    GetGraphRequest,
    GetGraphResponse,
    UpdateGraphRequest,
    UpdateGraphResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class GraphServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetFeatureSQL: UnaryUnaryMultiCallable[
        GetFeatureSQLRequest,
        GetFeatureSQLResponse,
    ]
    """GetFeatureSQL returns the feature SQLs for a given deployment."""
    GetFeaturesMetadata: UnaryUnaryMultiCallable[
        GetFeaturesMetadataRequest,
        GetFeaturesMetadataResponse,
    ]
    GetGraph: UnaryUnaryMultiCallable[
        GetGraphRequest,
        GetGraphResponse,
    ]
    UpdateGraph: UnaryUnaryMultiCallable[
        UpdateGraphRequest,
        UpdateGraphResponse,
    ]
    """UpdateGraph uploads the protobuf graph for a given deployment."""
    GetCodegenFeaturesFromGraph: UnaryUnaryMultiCallable[
        GetCodegenFeaturesFromGraphRequest,
        GetCodegenFeaturesFromGraphResponse,
    ]
    """GetCodegenFeaturesFromGraph returns Chalk features generated from the protograph"""

class GraphServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetFeatureSQL(
        self,
        request: GetFeatureSQLRequest,
        context: ServicerContext,
    ) -> GetFeatureSQLResponse:
        """GetFeatureSQL returns the feature SQLs for a given deployment."""
    @abstractmethod
    def GetFeaturesMetadata(
        self,
        request: GetFeaturesMetadataRequest,
        context: ServicerContext,
    ) -> GetFeaturesMetadataResponse: ...
    @abstractmethod
    def GetGraph(
        self,
        request: GetGraphRequest,
        context: ServicerContext,
    ) -> GetGraphResponse: ...
    @abstractmethod
    def UpdateGraph(
        self,
        request: UpdateGraphRequest,
        context: ServicerContext,
    ) -> UpdateGraphResponse:
        """UpdateGraph uploads the protobuf graph for a given deployment."""
    @abstractmethod
    def GetCodegenFeaturesFromGraph(
        self,
        request: GetCodegenFeaturesFromGraphRequest,
        context: ServicerContext,
    ) -> GetCodegenFeaturesFromGraphResponse:
        """GetCodegenFeaturesFromGraph returns Chalk features generated from the protograph"""

def add_GraphServiceServicer_to_server(servicer: GraphServiceServicer, server: Server) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/engine/v1/bloom_filter.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n"chalk/engine/v1/bloom_filter.proto\x12\x0f\x63halk.engine.v1"\xf2\x01\n\x0b\x42loomFilter\x12 \n\x0b\x65nvironment\x18\x01 \x01(\tR\x0b\x65nvironment\x12\x1c\n\tnamespace\x18\x02 \x01(\tR\tnamespace\x12\x1f\n\x0bnum_entries\x18\x03 \x01(\x04R\nnumEntries\x12\x30\n\x14num_expected_entries\x18\x04 \x01(\x04R\x12numExpectedEntries\x12\x1d\n\nnum_hashes\x18\x05 \x01(\x04R\tnumHashes\x12\x1d\n\nsize_bytes\x18\x06 \x01(\x04R\tsizeBytes\x12\x12\n\x04\x64\x61ta\x18\x07 \x01(\x0cR\x04\x64\x61taB\x85\x01\n\x13\x63om.chalk.engine.v1B\x10\x42loomFilterProtoP\x01\xa2\x02\x03\x43\x45X\xaa\x02\x0f\x43halk.Engine.V1\xca\x02\x0f\x43halk\\Engine\\V1\xe2\x02\x1b\x43halk\\Engine\\V1\\GPBMetadata\xea\x02\x11\x43halk::Engine::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.engine.v1.bloom_filter_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.engine.v1B\020BloomFilterProtoP\001\242\002\003CEX\252\002\017Chalk.Engine.V1\312\002\017Chalk\\Engine\\V1\342\002\033Chalk\\Engine\\V1\\GPBMetadata\352\002\021Chalk::Engine::V1"
    _globals["_BLOOMFILTER"]._serialized_start = 56
    _globals["_BLOOMFILTER"]._serialized_end = 298
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/common/v1/operation_kind.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b"\n$chalk/common/v1/operation_kind.proto\x12\x0f\x63halk.common.v1*\x94\x02\n\rOperationKind\x12\x1e\n\x1aOPERATION_KIND_UNSPECIFIED\x10\x00\x12\x1f\n\x1bOPERATION_KIND_ONLINE_QUERY\x10\x01\x12\x1c\n\x18OPERATION_KIND_MIGRATION\x10\x02\x12\x17\n\x13OPERATION_KIND_CRON\x10\x03\x12\x1c\n\x18OPERATION_KIND_STREAMING\x10\x04\x12$\n OPERATION_KIND_MIGRATION_SAMPLER\x10\x05\x12%\n!OPERATION_KIND_WINDOWED_STREAMING\x10\x06\x12 \n\x1cOPERATION_KIND_OFFLINE_QUERY\x10\x07\x42\x87\x01\n\x13\x63om.chalk.common.v1B\x12OperationKindProtoP\x01\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43halk.Common.V1\xca\x02\x0f\x43halk\\Common\\V1\xe2\x02\x1b\x43halk\\Common\\V1\\GPBMetadata\xea\x02\x11\x43halk::Common::V1b\x06proto3"
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.common.v1.operation_kind_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.common.v1B\022OperationKindProtoP\001\242\002\003CCX\252\002\017Chalk.Common.V1\312\002\017Chalk\\Common\\V1\342\002\033Chalk\\Common\\V1\\GPBMetadata\352\002\021Chalk::Common::V1"
    _globals["_OPERATIONKIND"]._serialized_start = 58
    _globals["_OPERATIONKIND"]._serialized_end = 334
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.named_query_pb2 import (
    GetAllNamedQueriesActiveDeploymentRequest,
    GetAllNamedQueriesActiveDeploymentResponse,
    GetAllNamedQueriesRequest,
    GetAllNamedQueriesResponse,
    GetNamedQueryByNameRequest,
    GetNamedQueryByNameResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class NamedQueryServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetAllNamedQueries: UnaryUnaryMultiCallable[
        GetAllNamedQueriesRequest,
        GetAllNamedQueriesResponse,
    ]
    GetAllNamedQueriesActiveDeployment: UnaryUnaryMultiCallable[
        GetAllNamedQueriesActiveDeploymentRequest,
        GetAllNamedQueriesActiveDeploymentResponse,
    ]
    GetNamedQueryByName: UnaryUnaryMultiCallable[
        GetNamedQueryByNameRequest,
        GetNamedQueryByNameResponse,
    ]

class NamedQueryServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetAllNamedQueries(
        self,
        request: GetAllNamedQueriesRequest,
        context: ServicerContext,
    ) -> GetAllNamedQueriesResponse: ...
    @abstractmethod
    def GetAllNamedQueriesActiveDeployment(
        self,
        request: GetAllNamedQueriesActiveDeploymentRequest,
        context: ServicerContext,
    ) -> GetAllNamedQueriesActiveDeploymentResponse: ...
    @abstractmethod
    def GetNamedQueryByName(
        self,
        request: GetNamedQueryByNameRequest,
        context: ServicerContext,
    ) -> GetNamedQueryByNameResponse: ...

def add_NamedQueryServiceServicer_to_server(servicer: NamedQueryServiceServicer, server: Server) -> None: ...



import datetime as dt
from typing import Any, Dict, Optional, Sequence, Tuple

from rich import box
from rich.jupyter import JupyterMixin
from rich.markdown import Markdown
from rich.panel import Panel
from rich.style import Style
from rich.table import Table

from chalk._reporting.rich.color import CHALK_WEBSITE_GREEN
from chalk.parsed.duplicate_input_gql import UpsertFeatureGQL, UpsertFilterGQL, UpsertResolverGQL
from chalk.utils.log_with_context import get_logger

_logger = get_logger(__name__)


class RichDisplayMixin:
    def _to_rich_obj(self) -> JupyterMixin:
        raise NotImplementedError()

    def _repr_mimebundle_(
        self,
        include: Sequence[str],
        exclude: Sequence[str],
        **kwargs: Any,
    ) -> Dict[str, str]:
        return self._to_rich_obj()._repr_mimebundle_(include, exclude, **kwargs)  # pyright: ignore


def get_default_table_with_panel(title: str) -> Tuple[Panel, Table]:
    table = Table.grid()
    panel = Panel.fit(
        table, border_style=Style(color=CHALK_WEBSITE_GREEN), box=box.ROUNDED, padding=(1, 2), title=title
    )
    return panel, table


def _print_timedelta(delta: dt.timedelta):
    if delta.days > 1:
        original_time = dt.datetime.now() - delta
        return "on " + original_time.strftime("%Y-%m-%d %H:%M").lower()
    elif delta.seconds >= 3600:
        return f"{delta.seconds // 3600}h{(delta.seconds % 3600) // 60}m ago"
    elif delta.seconds >= 60:
        return f"{delta.seconds // 60}m{delta.seconds % 60}s ago"
    elif delta.seconds > 1:
        return f"{delta.seconds} seconds ago"
    else:
        return "just now"


def _get_last_updated_markdown(last_update: Optional[dt.datetime]):
    if last_update is None:
        msg = "(from deployment source)"
    else:
        delta_text = _print_timedelta(dt.datetime.now(tz=dt.timezone.utc) - last_update)
        msg = f"(updated from notebook {delta_text})"
    return msg


class ResolverRichSummary(RichDisplayMixin):
    def __init__(self, gql: UpsertResolverGQL, last_update: Optional[dt.datetime]):
        super().__init__()
        self.gql: UpsertResolverGQL = gql
        self.last_update = last_update

    def _to_rich_obj(self) -> JupyterMixin:
        panel, table = get_default_table_with_panel(
            title=f"Resolver: {self.gql.fqn} {_get_last_updated_markdown(last_update=self.last_update)}"
        )
        # TODO detect if we need sql highlighting
        table.add_row(Markdown(f"""```python\n{self.gql.functionDefinition}\n```"""))

        return panel


class FeatureFieldRichSummary(RichDisplayMixin):
    def __init__(self, gql: UpsertFeatureGQL, last_update: Optional[dt.datetime]):
        super().__init__()
        self.gql: UpsertFeatureGQL = gql
        self.last_update = last_update

    def _to_rich_obj(self) -> JupyterMixin:
        panel, table = get_default_table_with_panel(
            title=f"Feature: {self.gql.id.fqn} {_get_last_updated_markdown(last_update=self.last_update)}"
        )
        table.add_row(
            Markdown(
                f"""**{self.gql.id.fqn}**\n
- _type:_\t{self._print_type()}\n
- _owner:_\t{self.gql.owner}\n
- _description:_\t{self.gql.description}\n
        """
            )
        )

        return panel

    def _print_type(self):
        # type_str = f"{self.gql.scalarKind}"
        # if self.gql.id.is_primary

        if kind := self.gql.scalarKind:
            return f"{kind.scalarKind}"
        elif kind := self.gql.featureTimeKind:
            return f"FeatureTime"
        elif kind := self.gql.hasOneKind:
            filt = self._print_filter(kind.join)
            return f"has_one: {filt}"
        elif kind := self.gql.hasManyKind:
            filt = self._print_filter(kind.join)
            return f"has_many: {filt}"

    def _print_filter(self, filter: UpsertFilterGQL) -> str:
        return f"{filter.lhs.fqn} {filter.op} {filter.rhs.fqn}"



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class ChalkKubernetesPersistentVolume(_message.Message):
    __slots__ = ("spec", "metrics")
    SPEC_FIELD_NUMBER: _ClassVar[int]
    METRICS_FIELD_NUMBER: _ClassVar[int]
    spec: ChalkKubernetesPersistentVolumeSpec
    metrics: ChalkKubernetesPersistentVolumeMetrics
    def __init__(
        self,
        spec: _Optional[_Union[ChalkKubernetesPersistentVolumeSpec, _Mapping]] = ...,
        metrics: _Optional[_Union[ChalkKubernetesPersistentVolumeMetrics, _Mapping]] = ...,
    ) -> None: ...

class ChalkKubernetesPersistentVolumeSpec(_message.Message):
    __slots__ = ("storage_class", "name", "access_modes", "capacity", "status", "reclaim_policy", "claim")
    STORAGE_CLASS_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ACCESS_MODES_FIELD_NUMBER: _ClassVar[int]
    CAPACITY_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    RECLAIM_POLICY_FIELD_NUMBER: _ClassVar[int]
    CLAIM_FIELD_NUMBER: _ClassVar[int]
    storage_class: str
    name: str
    access_modes: _containers.RepeatedScalarFieldContainer[str]
    capacity: str
    status: str
    reclaim_policy: str
    claim: str
    def __init__(
        self,
        storage_class: _Optional[str] = ...,
        name: _Optional[str] = ...,
        access_modes: _Optional[_Iterable[str]] = ...,
        capacity: _Optional[str] = ...,
        status: _Optional[str] = ...,
        reclaim_policy: _Optional[str] = ...,
        claim: _Optional[str] = ...,
    ) -> None: ...

class ChalkKubernetesPersistentVolumeMetrics(_message.Message):
    __slots__ = ("capacity_bytes", "used_bytes", "available_bytes")
    CAPACITY_BYTES_FIELD_NUMBER: _ClassVar[int]
    USED_BYTES_FIELD_NUMBER: _ClassVar[int]
    AVAILABLE_BYTES_FIELD_NUMBER: _ClassVar[int]
    capacity_bytes: float
    used_bytes: float
    available_bytes: float
    def __init__(
        self,
        capacity_bytes: _Optional[float] = ...,
        used_bytes: _Optional[float] = ...,
        available_bytes: _Optional[float] = ...,
    ) -> None: ...
